from flood_adapt.database_builder.database_builder import (
    Basins,
    ConfigModel,
    FootprintsOptions,
    GuiConfigModel,
    SpatialJoinModel,
    SviConfigModel,
    TideGaugeConfigModel,
    UnitSystems,
    create_database,
)

from .metrics_utils import (
    BuildingsInfographicModel,
    EventInfographicModel,
    HomesInfographicModel,
    ImpactCategoriesModel,
    MetricModel,
    RiskInfographicModel,
    RoadsInfographicModel,
)

__all__ = [
    "Basins",
    "ConfigModel",
    "FootprintsOptions",
    "GuiConfigModel",
    "SpatialJoinModel",
    "SviConfigModel",
    "TideGaugeConfigModel",
    "UnitSystems",
    "create_database",
    "BuildingsInfographicModel",
    "EventInfographicModel",
    "HomesInfographicModel",
    "MetricModel",
    "RiskInfographicModel",
    "RoadsInfographicModel",
    "ImpactCategoriesModel",
]



from flood_adapt.objects.forcing.discharge import (
    DischargeConstant,
    DischargeCSV,
    DischargeSynthetic,
)
from flood_adapt.objects.forcing.forcing import ForcingSource, ForcingType, IForcing
from flood_adapt.objects.forcing.meteo_handler import MeteoHandler
from flood_adapt.objects.forcing.rainfall import (
    RainfallConstant,
    RainfallCSV,
    RainfallMeteo,
    RainfallNetCDF,
    RainfallSynthetic,
    RainfallTrack,
)
from flood_adapt.objects.forcing.tide_gauge import TideGauge, TideGaugeSource
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.forcing.timeseries import (
    BlockTimeseries,
    CSVTimeseries,
    GaussianTimeseries,
    ScsTimeseries,
    Scstype,
    ShapeType,
    SyntheticTimeseries,
    TimeseriesFactory,
    TriangleTimeseries,
)
from flood_adapt.objects.forcing.waterlevels import (
    SurgeModel,
    TideModel,
    WaterlevelCSV,
    WaterlevelGauged,
    WaterlevelModel,
    WaterlevelSynthetic,
)
from flood_adapt.objects.forcing.wind import (
    WindConstant,
    WindCSV,
    WindMeteo,
    WindNetCDF,
    WindSynthetic,
    WindTrack,
)

__all__ = [
    # Forcing
    "IForcing",
    "ForcingSource",
    "ForcingType",
    # Timeseries
    "ShapeType",
    "Scstype",
    "CSVTimeseries",
    "TimeseriesFactory",
    "SyntheticTimeseries",
    "BlockTimeseries",
    "ScsTimeseries",
    "GaussianTimeseries",
    "TriangleTimeseries",
    # TimeFrame
    "TimeFrame",
    # Discharge
    "DischargeConstant",
    "DischargeCSV",
    "DischargeSynthetic",
    # Waterlevels
    "WaterlevelCSV",
    "WaterlevelGauged",
    "WaterlevelModel",
    "WaterlevelSynthetic",
    "TideModel",
    "SurgeModel",
    "TideGauge",
    "TideGaugeSource",
    # Rainfall
    "RainfallConstant",
    "RainfallCSV",
    "RainfallSynthetic",
    "RainfallMeteo",
    "RainfallNetCDF",
    "RainfallTrack",
    # Wind
    "WindConstant",
    "WindCSV",
    "WindSynthetic",
    "WindMeteo",
    "WindNetCDF",
    "WindTrack",
    # Other
    "MeteoHandler",
]



from enum import Enum
from pathlib import Path
from typing import ClassVar, Optional

import cht_observations.observation_stations as cht_station
import pandas as pd
import requests
from noaa_coops.station import COOPSAPIError
from pydantic import BaseModel, model_validator

from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.forcing.timeseries import CSVTimeseries

logger = FloodAdaptLogging.getLogger("TideGauge")


class TideGaugeSource(str, Enum):
    """The accepted input for the variable source in tide_gauge."""

    file = "file"
    noaa_coops = "noaa_coops"


class TideGauge(BaseModel):
    """The accepted input for the variable tide_gauge in Site.

    The obs_station is used for the download of tide gauge data, to be added to the hazard model as water level boundary condition.

    Attributes
    ----------
    name : Optional[int, str]
        Name of the tide gauge. Default is None.
    description : Optional[str]
        Description of the tide gauge. Default is "".
    source : TideGaugeSource
        Source of the tide gauge data.
    reference : str
        Reference of the tide gauge data. Should be defined in site.sfincs.water_level
    ID : Optional[int]
        ID of the tide gauge data. Default is None.
    file : Optional[Path]
        Only for file based tide gauges. Should be a path relative to the static folder. Default is None.
    lat : Optional[float]
        Latitude of the tide gauge data. Default is None.
    lon : Optional[float]
        Longitude of the tide gauge data. Default is None.
    units : us.UnitTypesLength
        Units of the water levels in the downloaded file. Default is us.UnitTypesLength.meters.

    """

    name: Optional[int | str] = None
    description: Optional[str] = ""
    source: TideGaugeSource
    reference: str
    ID: Optional[int] = None  # Attribute used to download from correct gauge
    file: Optional[Path] = None  # for locally stored data
    lat: Optional[float] = None
    lon: Optional[float] = None
    units: us.UnitTypesLength = (
        us.UnitTypesLength.meters
    )  # units of the water levels in the downloaded file

    _cached_data: ClassVar[dict[str, pd.DataFrame]] = {}

    @model_validator(mode="after")
    def validate_selection_type(self) -> "TideGauge":
        if self.source == TideGaugeSource.file and self.file is None:
            raise ValueError(
                "If `source` is 'file' a file path relative to the static folder should be provided with the attribute 'file'."
            )
        elif self.source == TideGaugeSource.noaa_coops and self.ID is None:
            raise ValueError(
                "If `source` is 'noaa_coops' the id of the station should be provided with the attribute 'ID'."
            )

        return self

    def get_waterlevels_in_time_frame(
        self,
        time: TimeFrame,
        out_path: Optional[Path] = None,
        units: us.UnitTypesLength = us.UnitTypesLength.meters,
    ) -> pd.DataFrame:
        """Download waterlevel data from NOAA station using station_id, start and stop time.

        Parameters
        ----------
        time : TimeFrame
            Time model with start and end time.
        tide_gauge : TideGauge
            Tide gauge model.
        out_path : Optional[Path], optional
            Path to save the data, by default None.
        units : us.UnitTypesLength, optional
            Unit of the waterlevel, by default us.UnitTypesLength.meters.

        Returns
        -------
        pd.DataFrame
            Dataframe with time as index and the waterlevel for each observation station as columns.
        """
        logger.info(f"Retrieving waterlevels for tide gauge {self.ID} for {time}")
        if self.file:
            gauge_data = self._read_imported_waterlevels(time=time, path=self.file)
        else:
            gauge_data = self._download_tide_gauge_data(time=time)

        if gauge_data is None:
            logger.warning(f"Could not retrieve waterlevels for tide gauge {self.ID}")
            return pd.DataFrame()

        gauge_data.columns = [f"waterlevel_{self.ID}"]
        gauge_data = gauge_data * us.UnitfulLength(value=1.0, units=self.units).convert(
            units
        )

        if out_path is not None:
            Path(out_path).parent.mkdir(parents=True, exist_ok=True)
            gauge_data.to_csv(Path(out_path))
        return gauge_data

    def _read_imported_waterlevels(self, time: TimeFrame, path: Path) -> pd.DataFrame:
        """Read waterlevels from an imported csv file.

        Parameters
        ----------
        path : Path
            Path to the csv file containing the waterlevel data. The csv file should have a column with the waterlevel data and a column with the time data.

        Returns
        -------
        pd.DataFrame
            Dataframe with time as index and the waterlevel for each observation station as columns.
            The data is sliced to the time range specified in the time model.
        """
        return CSVTimeseries.load_file(
            path=path, units=us.UnitfulLength(value=0, units=self.units)
        ).to_dataframe(time_frame=time)

    def _download_tide_gauge_data(self, time: TimeFrame) -> pd.DataFrame | None:
        """Download waterlevel data from NOAA station using station_id, start and stop time.

        Parameters
        ----------
        obs_point : ObsPointModel
            Observation point model.
        source : str
            Source of the data.

        Returns
        -------
        pd.DataFrame
            Dataframe with time as index and the waterlevel of the observation station as the column.
        None
            If the data could not be downloaded.
        """
        cache_key = f"{self.ID}_{time.start_time}_{time.end_time}"
        if cache_key in self.__class__._cached_data:
            logger.info("Tide gauge data retrieved from cache")
            return self.__class__._cached_data[cache_key]

        try:
            source_obj = cht_station.source(self.source.value)
            series = source_obj.get_data(
                id=self.ID,
                tstart=time.start_time,
                tstop=time.end_time,
                datum=self.reference,
            )
            index = pd.date_range(
                start=time.start_time,
                end=time.end_time,
                freq=time.time_step,
                name="time",
            )
            series = series.reindex(index, method="nearest")
            df = pd.DataFrame(data=series, index=index)

        except (COOPSAPIError, requests.JSONDecodeError) as e:
            logger.error(
                f"Could not download tide gauge data for station {self.ID}. {e}"
            )
            return None

        # Cache the result
        self.__class__._cached_data[cache_key] = df

        return df



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional, Union

import geopandas as gpd
import numpy as np

from flood_adapt.config.site import Site
from flood_adapt.dbs_classes.interface.element import AbstractDatabaseElement
from flood_adapt.dbs_classes.interface.static import IDbsStatic


class IDatabase(ABC):
    base_path: Path
    input_path: Path
    output_path: Path
    static_path: Path

    @property
    @abstractmethod
    def site(self) -> Site: ...

    @property
    @abstractmethod
    def static(self) -> IDbsStatic: ...

    @property
    @abstractmethod
    def events(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def scenarios(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def strategies(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def measures(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def projections(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def benefits(self) -> AbstractDatabaseElement: ...

    @abstractmethod
    def __init__(
        self, database_path: Union[str, os.PathLike], site_name: str
    ) -> None: ...

    @abstractmethod
    def get_outputs(self) -> dict[str, Any]:
        pass

    @abstractmethod
    def get_topobathy_path(self) -> str:
        pass

    @abstractmethod
    def get_index_path(self) -> str:
        pass

    @abstractmethod
    def get_depth_conversion(self) -> float:
        pass

    @abstractmethod
    def get_max_water_level(
        self, scenario_name: str, return_period: Optional[int] = None
    ) -> np.ndarray:
        pass

    @abstractmethod
    def get_building_footprints(self, scenario_name: str) -> gpd.GeoDataFrame:
        pass

    @abstractmethod
    def get_roads(self, scenario_name: str) -> gpd.GeoDataFrame:
        pass

    @abstractmethod
    def get_aggregation(self, scenario_name: str) -> dict[str, gpd.GeoDataFrame]:
        pass

    @abstractmethod
    def get_aggregation_benefits(
        self, benefit_name: str
    ) -> dict[str, gpd.GeoDataFrame]:
        pass

    @abstractmethod
    def get_object_list(self, object_type: str) -> dict[str, Any]:
        pass

    @abstractmethod
    def has_run_hazard(self, scenario_name: str) -> None:
        pass

    @abstractmethod
    def cleanup(self) -> None:
        pass



import logging
import math
import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Optional, Union

import geopandas as gpd
import hydromt_sfincs.utils as utils
import numpy as np
import pandas as pd
import plotly.express as px
import pyproj
import shapely
import xarray as xr
from cht_cyclones.tropical_cyclone import TropicalCyclone
from cht_tide.read_bca import SfincsBoundary
from cht_tide.tide_predict import predict
from hydromt_sfincs import SfincsModel as HydromtSfincsModel
from hydromt_sfincs.quadtree import QuadtreeGrid
from numpy import matlib
from shapely.affinity import translate

from flood_adapt.adapter.interface.hazard_adapter import IHazardAdapter
from flood_adapt.config.config import Settings
from flood_adapt.config.site import Site
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.path_builder import (
    ObjectDir,
    TopLevelDir,
    db_path,
)
from flood_adapt.misc.utils import cd, resolve_filepath
from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.events.events import Event, Mode, Template
from flood_adapt.objects.events.hurricane import TranslationModel
from flood_adapt.objects.events.synthetic import SyntheticEvent
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.discharge import (
    DischargeConstant,
    DischargeCSV,
    DischargeSynthetic,
)
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    ForcingType,
    IDischarge,
    IForcing,
    IRainfall,
    IWaterlevel,
    IWind,
)
from flood_adapt.objects.forcing.meteo_handler import MeteoHandler
from flood_adapt.objects.forcing.rainfall import (
    RainfallConstant,
    RainfallCSV,
    RainfallMeteo,
    RainfallNetCDF,
    RainfallSynthetic,
    RainfallTrack,
)
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.forcing.waterlevels import (
    WaterlevelCSV,
    WaterlevelGauged,
    WaterlevelModel,
    WaterlevelSynthetic,
)
from flood_adapt.objects.forcing.wind import (
    WindConstant,
    WindCSV,
    WindMeteo,
    WindNetCDF,
    WindSynthetic,
    WindTrack,
)
from flood_adapt.objects.measures.measures import (
    FloodWall,
    GreenInfrastructure,
    Measure,
    Pump,
)
from flood_adapt.objects.projections.projections import (
    PhysicalProjection,
    Projection,
)
from flood_adapt.objects.scenarios.scenarios import Scenario

logger = FloodAdaptLogging.getLogger("SfincsAdapter")


class SfincsAdapter(IHazardAdapter):
    """Adapter for the SFINCS model.

    This class is used to run the SFINCS model and process the results.

    Attributes
    ----------
    settings : SfincsModel
        The settings for the SFINCS model.
    """

    _site: Site
    _model: HydromtSfincsModel

    ###############
    ### PUBLIC ####
    ###############

    ### HAZARD ADAPTER METHODS ###
    def __init__(self, model_root: Path):
        """Load overland sfincs model based on a root directory.

        Parameters
        ----------
        model_root : Path
            Root directory of overland sfincs model.
        """
        self.settings = self.database.site.sfincs
        self.units = self.database.site.gui.units
        self._model = HydromtSfincsModel(
            root=model_root.resolve().as_posix(),
            mode="r",
            logger=self._setup_sfincs_logger(model_root),
        )
        self._model.read()

    def read(self, path: Path):
        """Read the sfincs model from the current model root."""
        if Path(self._model.root).resolve() != Path(path).resolve():
            self._model.set_root(root=path.as_posix(), mode="r")
        self._model.read()

    def write(self, path_out: Union[str, os.PathLike], overwrite: bool = True):
        """Write the sfincs model configuration to a directory."""
        root = self.get_model_root()
        if not isinstance(path_out, Path):
            path_out = Path(path_out).resolve()

        if not path_out.exists():
            path_out.mkdir(parents=True)

        if root != path_out:
            shutil.copytree(root, path_out, dirs_exist_ok=True)

        write_mode = "w+" if overwrite else "w"
        with cd(path_out):
            self._model.set_root(root=path_out.as_posix(), mode=write_mode)
            self._model.write()

    def close_files(self):
        """Close all open files and clean up file handles."""
        for _logger in [logger, self.sfincs_logger]:
            if hasattr(_logger, "handlers"):
                for handler in _logger.handlers:
                    if isinstance(handler, logging.FileHandler):
                        handler.close()
                        _logger.removeHandler(handler)

    def __enter__(self) -> "SfincsAdapter":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        self.close_files()
        return False

    def has_run(self, scenario: Scenario) -> bool:
        """Check if the model has been run."""
        return self.run_completed(scenario)

    def execute(self, path: Path, strict: bool = True) -> bool:
        """
        Run the sfincs executable in the specified path.

        Parameters
        ----------
        path : str
            Path to the simulation folder.
            Default is None, in which case the model root is used.
        strict : bool, optional
            True: raise an error if the model fails to run.
            False: log a warning.
            Default is True.

        Returns
        -------
        bool
            True if the model ran successfully, False otherwise.

        """
        sfincs_bin = Settings().sfincs_bin_path
        if not sfincs_bin or not sfincs_bin.exists():
            raise FileNotFoundError(
                f"SFINCS binary not found at {sfincs_bin}. Please check your settings."
            )

        with cd(path):
            logger.info(f"Running SFINCS in {path}")
            process = subprocess.run(
                sfincs_bin.as_posix(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            self.sfincs_logger.info(process.stdout)
            logger.debug(process.stdout)

        self._cleanup_simulation_folder(path)

        if process.returncode != 0:
            if Settings().delete_crashed_runs:
                # Remove all files in the simulation folder except for the log files
                for subdir, dirs, files in os.walk(path, topdown=False):
                    for file in files:
                        if not file.endswith(".log"):
                            os.remove(os.path.join(subdir, file))

                    if not os.listdir(subdir):
                        shutil.rmtree(subdir, ignore_errors=True)

            if strict:
                raise RuntimeError(f"SFINCS model failed to run in {path}.")
            else:
                logger.error(f"SFINCS model failed to run in {path}.")

        return process.returncode == 0

    def run(self, scenario: Scenario):
        """Run the whole workflow (Preprocess, process and postprocess) for a given scenario."""
        self._ensure_no_existing_forcings()
        event = self.database.events.get(scenario.event)

        if event.mode == Mode.risk:
            self._run_risk_scenario(scenario=scenario)
        else:
            self._run_single_event(scenario=scenario, event=event)

    def preprocess(self, scenario: Scenario, event: Event):
        """
        Preprocess the SFINCS model for a given scenario.

        Parameters
        ----------
        scenario : Scenario
            Scenario to preprocess.
        event : Event, optional
            Event to preprocess, by default None.
        """
        # I dont like this due to it being state based and might break if people use functions in the wrong order
        # Currently only used to pass projection + event stuff to WaterlevelModel

        sim_path = self._get_simulation_path(scenario=scenario, sub_event=event)
        sim_path.mkdir(parents=True, exist_ok=True)
        template_path = (
            self.database.static.get_overland_sfincs_model().get_model_root()
        )

        with SfincsAdapter(model_root=template_path) as model:
            model._load_scenario_objects(scenario, event)
            is_risk = "Probabilistic " if model._event_set is not None else ""
            logger.info(
                f"Preprocessing Scenario `{model._scenario.name}`: {is_risk}Event `{model._event.name}`, Strategy `{model._strategy.name}`, Projection `{model._projection.name}`"
            )
            # Write template model to output path and set it as the model root so focings can write to it
            model.set_timing(model._event.time)
            model.write(sim_path)

            # Event
            for forcing in model._event.get_forcings():
                model.add_forcing(forcing)

            if model.rainfall is not None:
                logger.info(
                    f"Adding event's rainfall multiplier: {model._event.rainfall_multiplier}"
                )
                model.rainfall *= model._event.rainfall_multiplier

            # Measures
            for measure in model._strategy.get_hazard_measures():
                model.add_measure(measure)

            # Projection
            model.add_projection(model._projection)

            # Output
            model.add_obs_points()

            # Save any changes made to disk as well
            model.write(path_out=sim_path)

    def process(self, scenario: Scenario, event: Event):
        if event.mode != Mode.single_event:
            raise ValueError(f"Unsupported event mode: {event.mode}.")

        sim_path = self._get_simulation_path(scenario=scenario, sub_event=event)
        logger.info(f"Running SFINCS for single event Scenario `{scenario.name}`")
        self.execute(sim_path)

    def postprocess(self, scenario: Scenario, event: Event):
        if event.mode != Mode.single_event:
            raise ValueError(f"Unsupported event mode: {event.mode}.")

        logger.info(f"Postprocessing SFINCS for Scenario `{scenario.name}`")
        if not self.sfincs_completed(
            self._get_simulation_path(scenario, sub_event=event)
        ):
            raise RuntimeError("SFINCS was not run successfully!")

        self.write_floodmap_geotiff(scenario)
        self.plot_wl_obs(scenario)
        self.write_water_level_map(scenario)

    def set_timing(self, time: TimeFrame):
        """Set model reference times."""
        logger.info(f"Setting timing for the SFINCS model: `{time}`")
        self._model.set_config("tref", time.start_time)
        self._model.set_config("tstart", time.start_time)
        self._model.set_config("tstop", time.end_time)

    def add_forcing(self, forcing: IForcing):
        """Get forcing data and add it."""
        if forcing is None:
            return

        logger.info(
            f"Adding {forcing.type.capitalize()}: {forcing.source.capitalize()}"
        )
        if isinstance(forcing, IRainfall):
            self._add_forcing_rain(forcing)
        elif isinstance(forcing, IWind):
            self._add_forcing_wind(forcing)
        elif isinstance(forcing, IDischarge):
            self._add_forcing_discharge(forcing)
        elif isinstance(forcing, IWaterlevel):
            self._add_forcing_waterlevels(forcing)
        else:
            logger.warning(
                f"Skipping unsupported forcing type {forcing.__class__.__name__}"
            )

    def add_measure(self, measure: Measure):
        """Get measure data and add it."""
        logger.info(
            f"Adding {measure.__class__.__name__.capitalize()} `{measure.name}`"
        )

        if isinstance(measure, FloodWall):
            self._add_measure_floodwall(measure)
        elif isinstance(measure, GreenInfrastructure):
            self._add_measure_greeninfra(measure)
        elif isinstance(measure, Pump):
            self._add_measure_pump(measure)
        else:
            logger.warning(
                f"Skipping unsupported measure type {measure.__class__.__name__}"
            )

    def add_projection(self, projection: Projection):
        """Get forcing data currently in the sfincs model and add the projection it."""
        logger.info(f"Adding Projection `{projection.name}`")
        phys_projection = projection.physical_projection

        if phys_projection.sea_level_rise:
            if self.waterlevels is not None:
                logger.info(
                    f"Adding projected sea level rise `{phys_projection.sea_level_rise}`"
                )
                self.waterlevels += phys_projection.sea_level_rise.convert(
                    us.UnitTypesLength.meters
                )

        if phys_projection.rainfall_multiplier:
            if self.rainfall is not None:
                logger.info(
                    f"Adding projected rainfall multiplier `{phys_projection.rainfall_multiplier}`"
                )
                self.rainfall *= phys_projection.rainfall_multiplier

    ### GETTERS ###
    def get_model_time(self) -> TimeFrame:
        t0, t1 = self._model.get_model_time()
        return TimeFrame(start_time=t0, end_time=t1)

    def get_model_root(self) -> Path:
        return Path(self._model.root)

    def get_mask(self):
        """Get mask with inactive cells from model."""
        mask = self._model.grid["msk"]
        return mask

    def get_bedlevel(self):
        """Get bed level from model."""
        self._model.read_results()
        zb = self._model.results["zb"]
        return zb

    def get_model_boundary(self) -> gpd.GeoDataFrame:
        """Get bounding box from model."""
        return self._model.region

    def get_model_grid(self) -> QuadtreeGrid:
        """Get grid from model.

        Returns
        -------
        QuadtreeGrid
            QuadtreeGrid with the model grid
        """
        return self._model.quadtree

    # Forcing properties
    @property
    def waterlevels(self) -> xr.Dataset | xr.DataArray | None:
        return self._model.forcing.get("bzs")

    @waterlevels.setter
    def waterlevels(self, waterlevels: xr.Dataset | xr.DataArray):
        if self.waterlevels is None or self.waterlevels.size == 0:
            raise ValueError("No water level forcing found in the model.")
        self._model.forcing["bzs"] = waterlevels

    @property
    def discharge(self) -> xr.Dataset | xr.DataArray | None:
        return self._model.forcing.get("dis")

    @discharge.setter
    def discharge(self, discharge: xr.Dataset | xr.DataArray):
        if self.discharge is None or self.discharge.size == 0:
            raise ValueError("No discharge forcing found in the model.")
        self._model.forcing["dis"] = discharge

    @property
    def rainfall(self) -> xr.Dataset | xr.DataArray | None:
        names = ["precip", "precip_2d"]
        in_model = [name for name in names if name in self._model.forcing]
        if len(in_model) == 0:
            return None
        elif len(in_model) == 1:
            return self._model.forcing[in_model[0]]
        else:
            raise ValueError("Multiple rainfall forcings found in the model.")

    @rainfall.setter
    def rainfall(self, rainfall: xr.Dataset | xr.DataArray):
        if self.rainfall is None or self.rainfall.size == 0:
            raise ValueError("No rainfall forcing found in the model.")
        elif "precip_2d" in self._model.forcing:
            self._model.forcing["precip_2d"] = rainfall
        elif "precip" in self._model.forcing:
            self._model.forcing["precip"] = rainfall
        else:
            raise ValueError("Unsupported rainfall forcing in the model.")

    @property
    def wind(self) -> xr.Dataset | xr.DataArray | None:
        wind_names = ["wnd", "wind_2d", "wind", "wind10_u", "wind10_v"]
        wind_in_model = [name for name in wind_names if name in self._model.forcing]
        if len(wind_in_model) == 0:
            return None
        elif len(wind_in_model) == 1:
            return self._model.forcing[wind_in_model[0]]
        elif len(wind_in_model) == 2:
            if not ("wind10_u" in wind_in_model and "wind10_v" in wind_in_model):
                raise ValueError(
                    "Multiple wind forcings found in the model. Both should be wind10_u and wind10_v or a singular wind forcing."
                )
            return xr.Dataset(
                {
                    "wind10_u": self._model.forcing["wind10_u"],
                    "wind10_v": self._model.forcing["wind10_v"],
                }
            )
        else:
            raise ValueError("Multiple wind forcings found in the model.")

    @wind.setter
    def wind(self, wind: xr.Dataset | xr.DataArray):
        if (not self.wind) or (self.wind.size == 0):
            raise ValueError("No wind forcing found in the model.")

        elif "wind_2d" in self._model.forcing:
            self._model.forcing["wind_2d"] = wind
        elif "wind" in self._model.forcing:
            self._model.forcing["wind"] = wind
        elif "wnd" in self._model.forcing:
            self._model.forcing["wnd"] = wind
        elif "wind10_u" in self._model.forcing and "wind10_v" in self._model.forcing:
            self._model.forcing["wind10_u"] = wind["wind10_u"]
            self._model.forcing["wind10_v"] = wind["wind10_v"]
        else:
            raise ValueError("Unsupported wind forcing in the model.")

    ### OUTPUT ###
    def run_completed(self, scenario: Scenario) -> bool:
        """Check if the entire model run has been completed successfully by checking if all flood maps exist that are created in postprocess().

        Returns
        -------
        bool : True if all flood maps exist, False otherwise.

        """
        paths = self._get_flood_map_paths(scenario)
        any_floodmap = len(paths) > 0
        all_exist = all(floodmap.exists() for floodmap in paths)
        return any_floodmap and all_exist

    def sfincs_completed(self, sim_path: Path) -> bool:
        """Check if the sfincs executable has been run successfully by checking if the output files exist in the simulation folder.

        Parameters
        ----------
        sim_path : Path
            Path to the simulation folder to check.

        Returns
        -------
        bool: True if the sfincs executable has been run successfully, False otherwise.

        """
        SFINCS_OUTPUT_FILES = ["sfincs_map.nc"]

        if self.settings.obs_point is not None:
            SFINCS_OUTPUT_FILES.append("sfincs_his.nc")

        to_check = [Path(sim_path) / file for file in SFINCS_OUTPUT_FILES]
        return all(output.exists() for output in to_check)

    def write_floodmap_geotiff(
        self, scenario: Scenario, sim_path: Optional[Path] = None
    ):
        """
        Read simulation results from SFINCS and saves a geotiff with the maximum water levels.

        Produced floodmap is in the units defined in the sfincs config settings.

        Parameters
        ----------
        scenario : Scenario
            Scenario for which to create the floodmap.
        sim_path : Path, optional
            Path to the simulation folder, by default None.
        """
        logger.info("Writing flood maps to geotiff")
        results_path = self._get_result_path(scenario)
        sim_path = sim_path or self._get_simulation_path(scenario)
        demfile = self.database.static_path / "dem" / self.settings.dem.filename

        with SfincsAdapter(model_root=sim_path) as model:
            zsmax = model._get_zsmax()

            dem = model._model.data_catalog.get_rasterdataset(demfile)

            # convert dem from dem units to floodmap units
            dem_conversion = us.UnitfulLength(
                value=1.0, units=self.settings.dem.units
            ).convert(self.settings.config.floodmap_units)

            floodmap_fn = results_path / f"FloodMap_{scenario.name}.tif"

            # convert zsmax from meters to floodmap units
            floodmap_conversion = us.UnitfulLength(
                value=1.0, units=us.UnitTypesLength.meters
            ).convert(self.settings.config.floodmap_units)

            utils.downscale_floodmap(
                zsmax=floodmap_conversion * zsmax,
                dep=dem_conversion * dem,
                hmin=0.01,
                floodmap_fn=floodmap_fn.as_posix(),
            )

    def write_water_level_map(
        self, scenario: Scenario, sim_path: Optional[Path] = None
    ):
        """Read simulation results from SFINCS and saves a netcdf with the maximum water levels."""
        logger.info("Writing water level map to netcdf")
        results_path = self._get_result_path(scenario)
        sim_path = sim_path or self._get_simulation_path(scenario)

        with SfincsAdapter(model_root=sim_path) as model:
            zsmax = model._get_zsmax()
            zsmax.to_netcdf(results_path / "max_water_level_map.nc")

    def plot_wl_obs(
        self,
        scenario: Scenario,
    ):
        """Plot water levels at SFINCS observation points as html.

        Only for single event scenarios, or for a specific simulation path containing the written and processed sfincs model.
        """
        if not self.settings.obs_point:
            logger.warning("No observation points provided in config.")
            return

        logger.info("Plotting water levels at observation points")
        sim_path = self._get_simulation_path(scenario)

        # read SFINCS model
        with SfincsAdapter(model_root=sim_path) as model:
            df, gdf = model._get_zs_points()

        gui_units = us.UnitTypesLength(
            self.database.site.gui.units.default_length_units
        )
        conversion_factor = us.UnitfulLength(
            value=1.0, units=us.UnitTypesLength("meters")
        ).convert(gui_units)

        overland_reference_height = self.settings.water_level.get_datum(
            self.settings.config.overland_model.reference
        ).height.convert(gui_units)

        for ii, col in enumerate(df.columns):
            # Plot actual thing
            fig = px.line(
                df[col] * conversion_factor
                + overland_reference_height  # convert to reference datum for plotting
            )

            fig.add_hline(
                y=0,
                line_dash="dash",
                line_color="#000000",
                annotation_text=self.settings.water_level.reference,
                annotation_position="bottom right",
            )

            # plot reference water levels
            for wl_ref in self.settings.water_level.datums:
                if (
                    wl_ref.name == self.settings.config.overland_model.reference
                    or wl_ref.name in self.database.site.gui.plotting.excluded_datums
                ):
                    continue
                fig.add_hline(
                    y=wl_ref.height.convert(gui_units),
                    line_dash="dash",
                    line_color="#3ec97c",
                    annotation_text=wl_ref.name,
                    annotation_position="bottom right",
                )

            fig.update_layout(
                autosize=False,
                height=100 * 2,
                width=280 * 2,
                margin={"r": 0, "l": 0, "b": 0, "t": 20},
                font={"size": 10, "color": "black", "family": "Arial"},
                title={
                    "text": gdf.iloc[ii]["Description"],
                    "font": {"size": 12, "color": "black", "family": "Arial"},
                    "x": 0.5,
                    "xanchor": "center",
                },
                xaxis_title="Time",
                yaxis_title=f"Water level [{gui_units.value}] above {self.settings.water_level.reference}",
                yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
                xaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
                showlegend=False,
            )

            event = self.database.events.get(scenario.event)
            if (
                self.settings.tide_gauge is not None
                and self.settings.obs_point[ii].name == self.settings.tide_gauge.name
            ):
                self._add_tide_gauge_plot(fig, event, units=gui_units)

            # write html to results folder
            station_name = gdf.iloc[ii]["Name"]
            results_path = self._get_result_path(scenario)
            fig.write_html(results_path / f"{station_name}_timeseries.html")

    def add_obs_points(self):
        """Add observation points provided in the site toml to SFINCS model."""
        obs_points = self.settings.obs_point
        if not obs_points:
            return

        names = [pt.name for pt in obs_points]
        lat = [pt.lat for pt in obs_points]
        lon = [pt.lon for pt in obs_points]

        logger.info("Adding observation points to the overland flood model")
        df = pd.DataFrame({"name": names})
        gdf = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(lon, lat),
            crs="EPSG:4326",
        )

        # Add locations to SFINCS file
        self._model.setup_observation_points(locations=gdf, merge=False)

    def get_wl_df_from_offshore_his_results(self) -> pd.DataFrame:
        """Create a pd.Dataframe with waterlevels from the offshore model at the bnd locations of the overland model.

        Returns
        -------
        wl_df: pd.DataFrame
            time series of water level.
        """
        logger.info("Reading water levels from offshore model")
        ds_his = utils.read_sfincs_his_results(
            Path(self._model.root) / "sfincs_his.nc",
            crs=self._model.crs.to_epsg(),
        )
        wl_df = pd.DataFrame(
            data=ds_his.point_zs.to_numpy(),
            index=ds_his.time.to_numpy(),
            columns=np.arange(1, ds_his.point_zs.to_numpy().shape[1] + 1, 1),
        )
        return wl_df

    ## RISK EVENTS ##
    def calculate_rp_floodmaps(self, scenario: Scenario):
        """Calculate flood risk maps from a set of (currently) SFINCS water level outputs using linear interpolation.

        It would be nice to make it more widely applicable and move the loading of the SFINCS results to self.postprocess_sfincs().

        generates return period water level maps in netcdf format to be used by FIAT
        generates return period water depth maps in geotiff format as product for users

        TODO: make this robust and more efficient for bigger datasets.
        """
        event: EventSet = self.database.events.get(scenario.event, load_all=True)
        if not isinstance(event, EventSet):
            raise ValueError("This function is only available for risk scenarios.")

        result_path = self._get_result_path(scenario)
        sim_paths = [
            self._get_simulation_path(scenario, sub_event=sub_event)
            for sub_event in event._events
        ]

        phys_proj = self.database.projections.get(
            scenario.projection
        ).physical_projection

        floodmap_rp = self.database.site.fiat.risk.return_periods
        frequencies = [sub_event.frequency for sub_event in event.sub_events]

        # adjust storm frequency for hurricane events
        if not math.isclose(phys_proj.storm_frequency_increase, 0, abs_tol=1e-9):
            storminess_increase = phys_proj.storm_frequency_increase / 100.0
            for ii, event in enumerate(event._events):
                if event.template == Template.Hurricane:
                    frequencies[ii] = frequencies[ii] * (1 + storminess_increase)

        with SfincsAdapter(model_root=sim_paths[0]) as dummymodel:
            # read mask and bed level
            mask = dummymodel.get_mask()
            zb = dummymodel.get_bedlevel()

        zs_maps = []
        for simulation_path in sim_paths:
            # read zsmax data from overland sfincs model
            with SfincsAdapter(model_root=simulation_path) as sim:
                zsmax = sim._get_zsmax().load()
                zs_maps.append(zsmax)

        # Create RP flood maps
        logger.info("Calculating flood risk maps, this may take some time")
        rp_flood_maps = self.calc_rp_maps(
            floodmaps=zs_maps,
            frequencies=frequencies,
            zb=zb,
            mask=mask,
            return_periods=floodmap_rp,
        )

        for ii, rp in enumerate(floodmap_rp):
            zs_rp_single = rp_flood_maps[ii]
            zs_rp_single = zs_rp_single.rio.write_crs(
                zsmax.raster.crs
            )  # , inplace=True)
            zs_rp_single = zs_rp_single.to_dataset(name="risk_map").transpose()
            fn_rp = result_path / f"RP_{rp:04d}_maps.nc"
            zs_rp_single.to_netcdf(fn_rp)

            # write geotiff
            # dem file for high resolution flood depth map
            demfile = self.database.static_path / "dem" / self.settings.dem.filename

            # writing the geotiff to the scenario results folder
            with SfincsAdapter(model_root=sim_paths[0]) as dummymodel:
                dem = dummymodel._model.data_catalog.get_rasterdataset(demfile)
                zsmax = zs_rp_single.to_array().squeeze().transpose()
                floodmap_fn = fn_rp.with_suffix(".tif")

                # convert dem from dem units to floodmap units
                dem_conversion = us.UnitfulLength(
                    value=1.0, units=self.settings.dem.units
                ).convert(self.settings.config.floodmap_units)

                # convert zsmax from meters to floodmap units
                floodmap_conversion = us.UnitfulLength(
                    value=1.0, units=us.UnitTypesLength.meters
                ).convert(self.settings.config.floodmap_units)

                utils.downscale_floodmap(
                    zsmax=floodmap_conversion * zsmax,
                    dep=dem_conversion * dem,
                    hmin=0.01,
                    floodmap_fn=floodmap_fn.as_posix(),
                )

    ######################################
    ### PRIVATE - use at your own risk ###
    ######################################
    def _run_single_event(self, scenario: Scenario, event: Event):
        self.preprocess(scenario, event)
        self.process(scenario, event)
        self.postprocess(scenario, event)

        if not self.settings.config.save_simulation:
            self._delete_simulation_folder(scenario, sub_event=event)

    def _delete_simulation_folder(
        self, scenario: Scenario, sub_event: Optional[Event] = None
    ):
        """Delete the simulation folder for a given scenario and optional sub-event."""
        sim_path = self._get_simulation_path(scenario, sub_event=sub_event)
        if sim_path.exists():
            shutil.rmtree(sim_path, ignore_errors=True)
            logger.info(f"Deleted simulation folder: {sim_path}")

        if sim_path.parent.exists() and not any(sim_path.parent.iterdir()):
            # Remove the parent directory `simulations` if it is empty
            shutil.rmtree(sim_path.parent, ignore_errors=True)

    def _run_risk_scenario(self, scenario: Scenario):
        """Run the whole workflow for a risk scenario.

        This means preprocessing and running the SFINCS model for each event in the event set, and then postprocessing the results.
        """
        event_set: EventSet = self.database.events.get(scenario.event, load_all=True)
        total = len(event_set._events)

        for i, sub_event in enumerate(event_set._events):
            sim_path = self._get_simulation_path(scenario, sub_event=sub_event)

            # Preprocess
            self.preprocess(scenario, event=sub_event)
            logger.info(
                f"Running SFINCS for Eventset Scenario `{scenario.name}`, Event `{sub_event.name}` ({i + 1}/{total})"
            )
            self.execute(sim_path)

        # Postprocess
        self.calculate_rp_floodmaps(scenario)

        # Cleanup
        if not self.settings.config.save_simulation:
            for i, sub_event in enumerate(event_set._events):
                shutil.rmtree(
                    self._get_simulation_path(scenario, sub_event=sub_event),
                    ignore_errors=True,
                )

    def _ensure_no_existing_forcings(self):
        """Check for existing forcings in the model and raise an error if any are found."""
        all_forcings = {
            "waterlevel": self.waterlevels,
            "rainfall": self.rainfall,
            "wind": self.wind,
            "discharge": self.discharge,
        }
        contains_forcings = ", ".join(
            [
                f"{name.capitalize()}"
                for name, forcing in all_forcings.items()
                if forcing is not None
            ]
        )
        if contains_forcings:
            raise ValueError(
                f"{contains_forcings} forcing(s) should not exists in the SFINCS template model. Remove it from the SFINCS model located at: {self.get_model_root()}. For more information on SFINCS and its input files, see the SFINCS documentation at: `https://sfincs.readthedocs.io/en/latest/input.html`"
            )

    ### FORCING ###
    def _add_forcing_wind(
        self,
        wind: IWind,
    ):
        """Add spatially constant wind forcing to sfincs model. Use timeseries or a constant magnitude and direction.

        Parameters
        ----------
        timeseries : Union[str, os.PathLike], optional
            path to file of timeseries file (.csv) which has three columns: time, magnitude and direction, by default None
        const_mag : float, optional
            magnitude of time-invariant wind forcing [m/s], by default None
        const_dir : float, optional
            direction of time-invariant wind forcing [deg], by default None
        """
        time_frame = self.get_model_time()
        if isinstance(wind, WindConstant):
            # HydroMT function: set wind forcing from constant magnitude and direction
            self._model.setup_wind_forcing(
                timeseries=None,
                magnitude=wind.speed.convert(us.UnitTypesVelocity.mps),
                direction=wind.direction.value,
            )
        elif isinstance(wind, WindSynthetic):
            df = wind.to_dataframe(time_frame=time_frame)
            df["mag"] *= us.UnitfulVelocity(
                value=1.0, units=self.units.default_velocity_units
            ).convert(us.UnitTypesVelocity.mps)

            tmp_path = Path(tempfile.gettempdir()) / "wind.csv"
            df.to_csv(tmp_path)

            # HydroMT function: set wind forcing from timeseries
            self._model.setup_wind_forcing(
                timeseries=tmp_path, magnitude=None, direction=None
            )
        elif isinstance(wind, WindMeteo):
            ds = MeteoHandler(
                dir=self.database.static_path / "meteo",
                lat=self.database.site.lat,
                lon=self.database.site.lon,
            ).read(time_frame)
            # data already in metric units so no conversion needed

            # HydroMT function: set wind forcing from grid
            self._model.setup_wind_forcing_from_grid(wind=ds)
        elif isinstance(wind, WindTrack):
            # data already in metric units so no conversion needed
            self._add_forcing_spw(wind)
        elif isinstance(wind, WindNetCDF):
            ds = wind.read()
            # time slicing to time_frame not needed, hydromt-sfincs handles it
            conversion = us.UnitfulVelocity(value=1.0, units=wind.units).convert(
                us.UnitTypesVelocity.mps
            )
            ds *= conversion
            self._model.setup_wind_forcing_from_grid(wind=ds)
        elif isinstance(wind, WindCSV):
            df = wind.to_dataframe(time_frame=time_frame)

            conversion = us.UnitfulVelocity(
                value=1.0, units=wind.units["speed"]
            ).convert(us.UnitTypesVelocity.mps)
            df *= conversion

            tmp_path = Path(tempfile.gettempdir()) / "wind.csv"
            df.to_csv(tmp_path)

            # HydroMT function: set wind forcing from timeseries
            self._model.setup_wind_forcing(
                timeseries=tmp_path,
                magnitude=None,
                direction=None,
            )
        else:
            logger.warning(f"Unsupported wind forcing type: {wind.__class__.__name__}")
            return

    def _add_forcing_rain(self, rainfall: IRainfall):
        """Add spatially constant rain forcing to sfincs model. Use timeseries or a constant magnitude.

        Parameters
        ----------
        timeseries : Union[str, os.PathLike], optional
            path to file of timeseries file (.csv) which has two columns: time and precipitation, by default None
        const_intensity : float, optional
            time-invariant precipitation intensity [mm_hr], by default None
        """
        time_frame = self.get_model_time()
        if isinstance(rainfall, RainfallConstant):
            self._model.setup_precip_forcing(
                timeseries=None,
                magnitude=rainfall.intensity.convert(us.UnitTypesIntensity.mm_hr),
            )
        elif isinstance(rainfall, RainfallCSV):
            df = rainfall.to_dataframe(time_frame=time_frame)
            conversion = us.UnitfulIntensity(value=1.0, units=rainfall.units).convert(
                us.UnitTypesIntensity.mm_hr
            )
            df *= conversion

            tmp_path = Path(tempfile.gettempdir()) / "precip.csv"
            df.to_csv(tmp_path)

            self._model.setup_precip_forcing(timeseries=tmp_path)
        elif isinstance(rainfall, RainfallSynthetic):
            df = rainfall.to_dataframe(time_frame=time_frame)

            if rainfall.timeseries.cumulative is not None:  # scs
                conversion = us.UnitfulLength(
                    value=1.0, units=rainfall.timeseries.cumulative.units
                ).convert(us.UnitTypesLength.millimeters)
            else:
                conversion = us.UnitfulIntensity(
                    value=1.0, units=rainfall.timeseries.peak_value.units
                ).convert(us.UnitTypesIntensity.mm_hr)

            df *= conversion
            tmp_path = Path(tempfile.gettempdir()) / "precip.csv"
            df.to_csv(tmp_path)

            self._model.setup_precip_forcing(timeseries=tmp_path)
        elif isinstance(rainfall, RainfallMeteo):
            ds = MeteoHandler(
                dir=self.database.static_path / "meteo",
                lat=self.database.site.lat,
                lon=self.database.site.lon,
            ).read(time_frame)
            # MeteoHandler always return metric so no conversion needed
            self._model.setup_precip_forcing_from_grid(precip=ds, aggregate=False)
        elif isinstance(rainfall, RainfallTrack):
            # data already in metric units so no conversion needed
            self._add_forcing_spw(rainfall)
        elif isinstance(rainfall, RainfallNetCDF):
            ds = rainfall.read()
            # time slicing to time_frame not needed, hydromt-sfincs handles it
            conversion = us.UnitfulIntensity(value=1.0, units=rainfall.units).convert(
                us.UnitTypesIntensity.mm_hr
            )
            ds *= conversion
            self._model.setup_precip_forcing_from_grid(precip=ds, aggregate=False)
        else:
            logger.warning(
                f"Unsupported rainfall forcing type: {rainfall.__class__.__name__}"
            )
            return

    def _add_forcing_discharge(self, forcing: IDischarge):
        """Add spatially constant discharge forcing to sfincs model. Use timeseries or a constant magnitude.

        Parameters
        ----------
        forcing : IDischarge
            The discharge forcing to add to the model.
            Can be a constant, synthetic or from a csv file.
            Also contains the river information.
        """
        if isinstance(forcing, (DischargeConstant, DischargeCSV, DischargeSynthetic)):
            self._set_single_river_forcing(discharge=forcing)
        else:
            logger.warning(
                f"Unsupported discharge forcing type: {forcing.__class__.__name__}"
            )

    def _add_forcing_waterlevels(self, forcing: IWaterlevel):
        time_frame = self.get_model_time()
        if isinstance(forcing, WaterlevelSynthetic):
            df_ts = forcing.to_dataframe(time_frame=time_frame)

            conversion = us.UnitfulLength(
                value=1.0, units=forcing.surge.timeseries.peak_value.units
            ).convert(us.UnitTypesLength.meters)
            datum_correction = self.settings.water_level.get_datum(
                self.database.site.gui.plotting.synthetic_tide.datum
            ).height.convert(us.UnitTypesLength.meters)

            df_ts = df_ts * conversion + datum_correction

            self._set_waterlevel_forcing(df_ts)
        elif isinstance(forcing, WaterlevelGauged):
            if self.settings.tide_gauge is None:
                raise ValueError("No tide gauge defined for this site.")

            df_ts = self.settings.tide_gauge.get_waterlevels_in_time_frame(
                time=time_frame,
            )
            conversion = us.UnitfulLength(
                value=1.0, units=self.settings.tide_gauge.units
            ).convert(us.UnitTypesLength.meters)

            datum_height = self.settings.water_level.get_datum(
                self.settings.tide_gauge.reference
            ).height.convert(us.UnitTypesLength.meters)

            df_ts = conversion * df_ts + datum_height

            self._set_waterlevel_forcing(df_ts)
        elif isinstance(forcing, WaterlevelCSV):
            df_ts = forcing.to_dataframe(time_frame=time_frame)

            if df_ts is None:
                raise ValueError("Failed to get waterlevel data.")
            conversion = us.UnitfulLength(value=1.0, units=forcing.units).convert(
                us.UnitTypesLength.meters
            )
            df_ts *= conversion
            self._set_waterlevel_forcing(df_ts)

        elif isinstance(forcing, WaterlevelModel):
            from flood_adapt.adapter.sfincs_offshore import OffshoreSfincsHandler

            if self.settings.config.offshore_model is None:
                raise ValueError("Offshore model configuration is missing.")
            if self._scenario is None or self._event is None:
                raise ValueError(
                    "Scenario and event must be provided to run the offshore model."
                )

            df_ts = OffshoreSfincsHandler(
                scenario=self._scenario, event=self._event
            ).get_resulting_waterlevels()
            if df_ts is None:
                raise ValueError("Failed to get waterlevel data.")

            # Datum
            datum_correction = self.settings.water_level.get_datum(
                self.settings.config.offshore_model.reference
            ).height.convert(us.UnitTypesLength.meters)
            df_ts += datum_correction

            # Already in meters since it was produced by SFINCS so no conversion needed
            self._set_waterlevel_forcing(df_ts)
            self._turn_off_bnd_press_correction()
        else:
            logger.warning(
                f"Unsupported waterlevel forcing type: {forcing.__class__.__name__}"
            )

    # SPIDERWEB
    def _add_forcing_spw(self, forcing: Union[RainfallTrack, WindTrack]):
        """Add spiderweb forcing."""
        if forcing.source != ForcingSource.TRACK:
            raise ValueError("Forcing source should be TRACK.")

        if forcing.path is None:
            raise ValueError("No path to track file provided.")

        if not forcing.path.exists():
            # Check if the file is in the database
            in_db = self._get_event_input_path(self._event) / forcing.path.name
            if not in_db.exists():
                raise FileNotFoundError(
                    f"Input file for track forcing not found: {forcing.path}"
                )
            forcing.path = in_db

        if forcing.path.suffix == ".cyc":
            forcing.path = self._create_spw_file_from_track(
                track_forcing=forcing,
                hurricane_translation=self._event.hurricane_translation,
                name=self._event.name,
                output_dir=forcing.path.parent,
                include_rainfall=bool(self._event.forcings.get(ForcingType.RAINFALL)),
                recreate=False,
            )

        if forcing.path.suffix != ".spw":
            raise ValueError(
                "Track files should be in one of [spw, ddb_cyc] file format and must have [.spw, .cyc] extension."
            )

        sim_path = self.get_model_root()
        logger.info(f"Adding spiderweb forcing to Sfincs model: {sim_path.name}")

        # prevent SameFileError
        output_spw_path = sim_path / forcing.path.name
        if forcing.path == output_spw_path:
            raise ValueError(
                "Add a different SPW file than the one already in the model."
            )

        if output_spw_path.exists():
            os.remove(output_spw_path)
        shutil.copy2(forcing.path, output_spw_path)

        self._model.set_config("spwfile", output_spw_path.name)

    ### MEASURES ###
    def _add_measure_floodwall(self, floodwall: FloodWall):
        """Add floodwall to sfincs model.

        Parameters
        ----------
        floodwall : FloodWall
            floodwall information
        """
        polygon_file = resolve_filepath(
            object_dir=ObjectDir.measure,
            obj_name=floodwall.name,
            path=floodwall.polygon_file,
        )

        # HydroMT function: get geodataframe from filename
        gdf_floodwall = self._model.data_catalog.get_geodataframe(
            polygon_file, geom=self._model.region, crs=self._model.crs
        )

        # Add floodwall attributes to geodataframe
        gdf_floodwall["name"] = floodwall.name
        if (gdf_floodwall.geometry.type == "MultiLineString").any():
            gdf_floodwall = gdf_floodwall.explode()

        try:
            heights = [
                float(
                    us.UnitfulLength(
                        value=float(height),
                        units=self.database.site.gui.units.default_length_units,
                    ).convert(us.UnitTypesLength("meters"))
                )
                for height in gdf_floodwall["z"]
            ]
            gdf_floodwall["z"] = heights
            logger.info("Using floodwall height from shape file.")
        except Exception:
            logger.warning(
                f"Could not use height data from file due to missing `z` column or missing values therein. Using uniform height of {floodwall.elevation} instead."
            )
            gdf_floodwall["z"] = floodwall.elevation.convert(
                us.UnitTypesLength(us.UnitTypesLength.meters)
            )

        # par1 is the overflow coefficient for weirs
        gdf_floodwall["par1"] = 0.6

        # HydroMT function: create floodwall
        self._model.setup_structures(structures=gdf_floodwall, stype="weir", merge=True)

    def _add_measure_greeninfra(self, green_infrastructure: GreenInfrastructure):
        # HydroMT function: get geodataframe from filename
        if green_infrastructure.selection_type == "polygon":
            polygon_file = resolve_filepath(
                ObjectDir.measure,
                green_infrastructure.name,
                green_infrastructure.polygon_file,
            )
        elif green_infrastructure.selection_type == "aggregation_area":
            # TODO this logic already exists in the Database controller but cannot be used due to cyclic imports
            # Loop through available aggregation area types
            for aggr_dict in self.database.site.fiat.config.aggregation:
                # check which one is used in measure
                if not aggr_dict.name == green_infrastructure.aggregation_area_type:
                    continue
                # load geodataframe
                aggr_areas = gpd.read_file(
                    db_path(TopLevelDir.static) / aggr_dict.file,
                    engine="pyogrio",
                ).to_crs(4326)
                # keep only aggregation area chosen
                polygon_file = aggr_areas.loc[
                    aggr_areas[aggr_dict.field_name]
                    == green_infrastructure.aggregation_area_name,
                    ["geometry"],
                ].reset_index(drop=True)
        else:
            raise ValueError(
                f"The selection type: {green_infrastructure.selection_type} is not valid"
            )

        gdf_green_infra = self._model.data_catalog.get_geodataframe(
            polygon_file,
            geom=self._model.region,
            crs=self._model.crs,
        )

        # Make sure no multipolygons are there
        gdf_green_infra = gdf_green_infra.explode()

        # HydroMT function: create storage volume
        self._model.setup_storage_volume(
            storage_locs=gdf_green_infra,
            volume=green_infrastructure.volume.convert(us.UnitTypesVolume.m3),
            merge=True,
        )

    def _add_measure_pump(self, pump: Pump):
        """Add pump to sfincs model.

        Parameters
        ----------
        pump : Pump
            pump information
        """
        polygon_file = resolve_filepath(ObjectDir.measure, pump.name, pump.polygon_file)
        # HydroMT function: get geodataframe from filename
        gdf_pump = self._model.data_catalog.get_geodataframe(
            polygon_file, geom=self._model.region, crs=self._model.crs
        )

        # HydroMT function: create floodwall
        self._model.setup_drainage_structures(
            structures=gdf_pump,
            stype="pump",
            discharge=pump.discharge.convert(us.UnitTypesDischarge.cms),
            merge=True,
        )

    ### SFINCS SETTERS ###
    def _set_single_river_forcing(self, discharge: IDischarge):
        """Add discharge to overland sfincs model.

        Parameters
        ----------
        discharge : IDischarge
            Discharge object with discharge timeseries data and river information.
        """
        if not isinstance(
            discharge, (DischargeConstant, DischargeSynthetic, DischargeCSV)
        ):
            logger.warning(
                f"Unsupported discharge forcing type: {discharge.__class__.__name__}"
            )
            return

        model_rivers = self._read_river_locations()
        if model_rivers.empty:
            logger.warning(
                "Cannot add discharge forcing: No rivers defined in the sfincs model."
            )
            return
        logger.info(f"Setting discharge forcing for river: {discharge.river.name}")
        time_frame = self.get_model_time()

        # Check that the river is defined in the model and that the coordinates match
        river_loc = shapely.Point(
            discharge.river.x_coordinate, discharge.river.y_coordinate
        )
        tolerance = 0.001  # in degrees, ~111 meters at the equator. (0.0001: 11 meters at the equator)
        river_gdf = model_rivers[model_rivers.distance(river_loc) <= tolerance]
        river_inds = river_gdf.index.to_list()
        if len(river_inds) != 1:
            raise ValueError(
                f"River {discharge.river.name} is not defined in the sfincs model. Please ensure the river coordinates in the site.toml match the coordinates for rivers in the SFINCS model."
            )

        # Create a geodataframe with the river coordinates, the timeseries data and rename the column to the river index defined in the model
        if isinstance(discharge, DischargeCSV):
            df = discharge.to_dataframe(time_frame)
            conversion = us.UnitfulDischarge(value=1.0, units=discharge.units).convert(
                us.UnitTypesDischarge.cms
            )
        elif isinstance(discharge, DischargeConstant):
            df = discharge.to_dataframe(time_frame)
            conversion = us.UnitfulDischarge(
                value=1.0, units=discharge.discharge.units
            ).convert(us.UnitTypesDischarge.cms)
        elif isinstance(discharge, DischargeSynthetic):
            df = discharge.to_dataframe(time_frame)
            conversion = us.UnitfulDischarge(
                value=1.0, units=discharge.timeseries.peak_value.units
            ).convert(us.UnitTypesDischarge.cms)
        else:
            raise ValueError(
                f"Unsupported discharge forcing type: {discharge.__class__}"
            )

        df *= conversion

        df = df.rename(columns={df.columns[0]: river_inds[0]})

        # HydroMT function: set discharge forcing from time series and river coordinates
        self._model.setup_discharge_forcing(
            locations=river_gdf,
            timeseries=df,
            merge=True,
        )

    def _turn_off_bnd_press_correction(self):
        """Turn off the boundary pressure correction in the sfincs model."""
        logger.info("Turning off boundary pressure correction in the offshore model")
        self._model.set_config("pavbnd", -9999)

    def _set_waterlevel_forcing(self, df_ts: pd.DataFrame):
        """
        Add water level forcing to sfincs model.

        Values in the timeseries are expected to be relative to the main reference datum: `self.settings.water_level.reference`.
        The overland model reference: `self.settings.config.overland_model.reference` is used to convert the water levels to the reference of the overland model.

        Parameters
        ----------
        df_ts : pd.DataFrame
            Time series of water levels with the first column as the time index.


        """
        # Determine bnd points from reference overland model
        gdf_locs = self._read_waterlevel_boundary_locations()

        if len(df_ts.columns) == 1:
            # Go from 1 timeseries to timeseries for all boundary points
            name = df_ts.columns[0]
            for i in range(1, len(gdf_locs)):
                df_ts[i + 1] = df_ts[name]
            df_ts.columns = list(range(1, len(gdf_locs) + 1))

        # Datum
        sfincs_overland_reference_height = self.settings.water_level.get_datum(
            self.settings.config.overland_model.reference
        ).height.convert(us.UnitTypesLength.meters)

        df_ts -= sfincs_overland_reference_height

        # HydroMT function: set waterlevel forcing from time series
        self._model.set_forcing_1d(
            name="bzs", df_ts=df_ts, gdf_locs=gdf_locs, merge=False
        )

    # OFFSHORE
    def _add_pressure_forcing_from_grid(self, ds: xr.DataArray):
        """Add spatially varying barometric pressure to sfincs model.

        Parameters
        ----------
        ds : xr.DataArray
            - Required variables: ['press_msl' (Pa)]
            - Required coordinates: ['time', 'y', 'x']
            - spatial_ref: CRS
        """
        logger.info("Adding pressure forcing to the offshore model")
        self._model.setup_pressure_forcing_from_grid(press=ds)

    def _add_bzs_from_bca(self, event: Event, physical_projection: PhysicalProjection):
        # ONLY offshore models
        """Convert tidal constituents from bca file to waterlevel timeseries that can be read in by hydromt_sfincs."""
        if self.settings.config.offshore_model is None:
            raise ValueError("No offshore model found in sfincs config.")

        logger.info("Adding water level forcing to the offshore model")
        sb = SfincsBoundary()
        sb.read_flow_boundary_points(self.get_model_root() / "sfincs.bnd")
        sb.read_astro_boundary_conditions(self.get_model_root() / "sfincs.bca")

        times = pd.date_range(
            start=event.time.start_time,
            end=event.time.end_time,
            freq="10T",
        )

        # Predict tidal signal and add SLR
        if not sb.flow_boundary_points:
            raise ValueError("No flow boundary points found.")

        if self.settings.config.offshore_model.vertical_offset:
            correction = self.settings.config.offshore_model.vertical_offset.convert(
                us.UnitTypesLength.meters
            )
        else:
            correction = 0.0

        for bnd_ii in range(len(sb.flow_boundary_points)):
            tide_ii = (
                predict(sb.flow_boundary_points[bnd_ii].astro, times)
                + correction
                + physical_projection.sea_level_rise.convert(us.UnitTypesLength.meters)
            )

            if bnd_ii == 0:
                wl_df = pd.DataFrame(data={1: tide_ii}, index=times)
            else:
                wl_df[bnd_ii + 1] = tide_ii

        # Determine bnd points from reference overland model
        gdf_locs = self._read_waterlevel_boundary_locations()

        # HydroMT function: set waterlevel forcing from time series
        self._model.set_forcing_1d(
            name="bzs", df_ts=wl_df, gdf_locs=gdf_locs, merge=False
        )

    ### PRIVATE GETTERS ###
    def _get_result_path(self, scenario: Scenario) -> Path:
        """Return the path to store the results."""
        return self.database.scenarios.output_path / scenario.name / "Flooding"

    def _get_simulation_path(
        self, scenario: Scenario, sub_event: Optional[Event] = None
    ) -> Path:
        """
        Return the path to the simulation results.

        Parameters
        ----------
        scenario : Scenario
            The scenario for which to get the simulation path.
        sub_event : Optional[Event], optional
            The sub-event for which to get the simulation path, by default None.
            Is only used when the event associated with the scenario is an EventSet.
        """
        base_path = (
            self._get_result_path(scenario)
            / "simulations"
            / self.settings.config.overland_model.name
        )
        event = self.database.events.get(scenario.event)

        if isinstance(event, EventSet):
            if sub_event is None:
                raise ValueError("Event must be provided when scenario is an EventSet.")
            return base_path.parent / sub_event.name / base_path.name
        elif isinstance(event, Event):
            return base_path
        else:
            raise ValueError(f"Unsupported mode: {event.mode}")

    def _get_simulation_path_offshore(
        self, scenario: Scenario, sub_event: Optional[Event] = None
    ) -> Path:
        # Get the path to the offshore model (will not be used if offshore model is not created)
        if self.settings.config.offshore_model is None:
            raise ValueError("No offshore model found in sfincs config.")
        base_path = (
            self._get_result_path(scenario)
            / "simulations"
            / self.settings.config.offshore_model.name
        )
        event = self.database.events.get(scenario.event)
        if isinstance(event, EventSet):
            return base_path.parent / sub_event.name / base_path.name
        elif isinstance(event, Event):
            return base_path
        else:
            raise ValueError(f"Unsupported mode: {event.mode}")

    def _get_flood_map_paths(self, scenario: Scenario) -> list[Path]:
        """Return the paths to the flood maps that running this scenario should produce."""
        results_path = self._get_result_path(scenario)
        event = self.database.events.get(scenario.event)

        if isinstance(event, EventSet):
            map_fn = []
            for rp in self.database.site.fiat.risk.return_periods:
                map_fn.append(results_path / f"RP_{rp:04d}_maps.nc")
        elif isinstance(event, Event):
            map_fn = [results_path / "max_water_level_map.nc"]
        else:
            raise ValueError(f"Unsupported mode: {event.mode}")

        return map_fn

    def _get_event_input_path(self, event: Event) -> Path:
        """Return the path to the event input directory."""
        return self.database.events.input_path / event.name

    def _get_zsmax(self):
        """Read zsmax file and return absolute maximum water level over entire simulation."""
        self._model.read_results()
        zsmax = self._model.results["zsmax"].max(dim="timemax")
        zsmax.attrs["units"] = "m"
        return zsmax

    def _get_zs_points(self):
        """Read water level (zs) timeseries at observation points.

        Names are allocated from the site.toml.
        See also add_obs_points() above.
        """
        self._model.read_results()
        da = self._model.results["point_zs"]
        df = pd.DataFrame(index=pd.DatetimeIndex(da.time), data=da.to_numpy())

        names = []
        descriptions = []
        # get station names from site.toml
        if self.settings.obs_point is not None:
            obs_points = self.settings.obs_point
            for pt in obs_points:
                names.append(pt.name)
                descriptions.append(pt.description)

        pt_df = pd.DataFrame({"Name": names, "Description": descriptions})
        gdf = gpd.GeoDataFrame(
            pt_df,
            geometry=gpd.points_from_xy(da.point_x.values, da.point_y.values),
            crs=self._model.crs,
        )
        return df, gdf

    def _create_spw_file_from_track(
        self,
        track_forcing: Union[RainfallTrack, WindTrack],
        hurricane_translation: TranslationModel,
        name: str,
        output_dir: Path,
        include_rainfall: bool = False,
        recreate: bool = False,
    ):
        """
        Create a spiderweb file from a given TropicalCyclone track and save it to the event's input directory.

        Providing the output_dir argument allows to save the spiderweb file in a different directory.

        Parameters
        ----------
        output_dir : Path
            The directory where the spiderweb file is saved (or copied to if it already exists and recreate is False)
        recreate : bool, optional
            If True, the spiderweb file is recreated even if it already exists, by default False

        Returns
        -------
        Path
            the path to the created spiderweb file
        """
        if track_forcing.path is None:
            raise ValueError("No path to track file provided.")

        # Check file format
        match track_forcing.path.suffix:
            case ".spw":
                if recreate:
                    raise ValueError(
                        "Recreating spiderweb files from existing spiderweb files is not supported. Provide a track file instead."
                    )

                if track_forcing.path.exists():
                    return track_forcing.path

                elif (output_dir / track_forcing.path.name).exists():
                    return output_dir / track_forcing.path.name

                else:
                    raise FileNotFoundError(f"SPW file not found: {track_forcing.path}")
            case ".cyc":
                pass
            case _:
                raise ValueError(
                    "Track files should be in the DDB_CYC file format and must have .cyc extension, or in the SPW file format and must have .spw extension"
                )

        # Check if the spiderweb file already exists
        spw_file = output_dir / track_forcing.path.with_suffix(".spw").name
        if spw_file.exists():
            if recreate:
                os.remove(spw_file)
            else:
                return spw_file

        # Initialize the tropical cyclone
        tc = TropicalCyclone()
        tc.read_track(filename=track_forcing.path.as_posix(), fmt="ddb_cyc")

        # Alter the track of the tc if necessary
        tc = self._translate_tc_track(
            tc=tc, hurricane_translation=hurricane_translation
        )

        # Rainfall
        start = "Including" if include_rainfall else "Excluding"
        logger.info(f"{start} rainfall in the spiderweb file")
        tc.include_rainfall = include_rainfall

        logger.info(
            f"Creating spiderweb file for hurricane event `{name}`. This may take a while."
        )

        # Create spiderweb file from the track
        tc.to_spiderweb(spw_file)

        return spw_file

    def _translate_tc_track(
        self, tc: TropicalCyclone, hurricane_translation: TranslationModel
    ):
        if math.isclose(
            hurricane_translation.eastwest_translation.value, 0, abs_tol=1e-6
        ) and math.isclose(
            hurricane_translation.northsouth_translation.value, 0, abs_tol=1e-6
        ):
            return tc

        logger.info(f"Translating the track of the tropical cyclone `{tc.name}`")
        # First convert geodataframe to the local coordinate system
        crs = pyproj.CRS.from_string(self.settings.config.csname)
        tc.track = tc.track.to_crs(crs)

        # Translate the track in the local coordinate system
        tc.track["geometry"] = tc.track["geometry"].apply(
            lambda geom: translate(
                geom,
                xoff=hurricane_translation.eastwest_translation.convert(
                    us.UnitTypesLength.meters
                ),
                yoff=hurricane_translation.northsouth_translation.convert(
                    us.UnitTypesLength.meters
                ),
            )
        )

        # Convert the geodataframe to lat/lon
        tc.track = tc.track.to_crs(epsg=4326)

        return tc

    # @gundula do we keep this func, its not used anywhere?
    def _downscale_hmax(self, zsmax, demfile: Path):
        # read DEM and convert units to metric units used by SFINCS
        demfile_units = self.settings.dem.units
        dem_conversion = us.UnitfulLength(value=1.0, units=demfile_units).convert(
            us.UnitTypesLength("meters")
        )
        dem = dem_conversion * self._model.data_catalog.get_rasterdataset(demfile)
        dem = dem.rio.reproject(self._model.crs)

        # determine conversion factor for output floodmap
        floodmap_units = self.settings.config.floodmap_units
        floodmap_conversion = us.UnitfulLength(
            value=1.0, units=us.UnitTypesLength.meters
        ).convert(floodmap_units)

        hmax = utils.downscale_floodmap(
            zsmax=floodmap_conversion * zsmax,
            dep=floodmap_conversion * dem,
            hmin=0.01,
        )
        return hmax

    def _read_river_locations(self) -> gpd.GeoDataFrame:
        path = self.get_model_root() / "sfincs.src"
        lines = []
        if path.exists():
            with open(path) as f:
                lines = f.readlines()
        coords = [(float(line.split()[0]), float(line.split()[1])) for line in lines]
        points = [shapely.Point(coord) for coord in coords]

        return gpd.GeoDataFrame({"geometry": points}, crs=self._model.crs)

    def _read_waterlevel_boundary_locations(self) -> gpd.GeoDataFrame:
        path = self.get_model_root() / "sfincs.bnd"
        lines = []
        if path.exists():
            with open(path) as f:
                lines = f.readlines()

        coords = [(float(line.split()[0]), float(line.split()[1])) for line in lines]
        points = [shapely.Point(coord) for coord in coords]

        return gpd.GeoDataFrame({"geometry": points}, crs=self._model.crs)

    def _setup_sfincs_logger(self, model_root: Path) -> logging.Logger:
        """Initialize the logger for the SFINCS model."""
        # Create a logger for the SFINCS model manually
        sfincs_logger = logging.getLogger("SfincsModel")
        for handler in sfincs_logger.handlers[:]:
            sfincs_logger.removeHandler(handler)

        # Add a file handler
        file_handler = logging.FileHandler(
            filename=model_root.resolve() / "sfincs_model.log",
            mode="w",
        )
        sfincs_logger.setLevel(logging.DEBUG)
        sfincs_logger.addHandler(file_handler)
        self.sfincs_logger = sfincs_logger
        return sfincs_logger

    def _cleanup_simulation_folder(
        self,
        path: Path,
        extensions: list[str] = [".spw"],
    ):
        """Remove all files with the given extensions in the given path."""
        if not path.exists():
            return

        for ext in extensions:
            for file in path.glob(f"*{ext}"):
                file.unlink()

    def _load_scenario_objects(self, scenario: Scenario, event: Event) -> None:
        self._scenario = scenario
        self._projection = self.database.projections.get(scenario.projection)
        self._strategy = self.database.strategies.get(scenario.strategy)
        self._event = event

        _event = self.database.events.get(scenario.event)
        if isinstance(_event, EventSet):
            self._event_set = _event
        else:
            self._event_set = None

    def _add_tide_gauge_plot(
        self, fig, event: Event, units: us.UnitTypesLength
    ) -> None:
        if isinstance(event, SyntheticEvent):
            return
        if self.settings.tide_gauge is None:
            return
        df_gauge = self.settings.tide_gauge.get_waterlevels_in_time_frame(
            time=TimeFrame(
                start_time=event.time.start_time,
                end_time=event.time.end_time,
            ),
            units=us.UnitTypesLength(units),
        )

        if df_gauge is None or df_gauge.empty:
            logger.warning(
                "No water level data available for the tide gauge. Could not add it to the plot."
            )
            return

        gauge_reference_height = self.settings.water_level.get_datum(
            self.settings.tide_gauge.reference
        ).height.convert(units)

        waterlevel = df_gauge.iloc[:, 0] + gauge_reference_height

        # If data is available, add to plot
        fig.add_trace(px.line(waterlevel, color_discrete_sequence=["#ea6404"]).data[0])
        fig["data"][0]["name"] = "model"
        fig["data"][1]["name"] = "measurement"
        fig.update_layout(showlegend=True)

    @staticmethod
    def calc_rp_maps(
        floodmaps: list[xr.DataArray],
        frequencies: list[float],
        zb: xr.DataArray,
        mask: xr.DataArray,
        return_periods: list[float],
    ) -> list[xr.DataArray]:
        """
        Calculate return period (RP) flood maps from a set of flood simulation results.

        This function processes multiple flood simulation outputs (water level maps) and their associated frequencies
        to generate hazard maps for specified return periods. It interpolates water levels for each return period
        using exceedance probabilities and handles masked or dry cells appropriately.

        Args:
            floodmaps (list[xr.DataArray]): List of water level maps (xarray DataArrays), one for each simulation.
            frequencies (list[float]): List of frequencies (probabilities of occurrence) corresponding to each floodmap.
            zb (np.ndarray): Array of bed elevations for each grid cell.
            mask (xr.DataArray): Mask indicating valid (1) and invalid (0) grid cells.
            return_periods (list[float]): List of return periods (in years) for which to generate hazard maps.

        Returns
        -------
            list[xr.DataArray]: List of xarray DataArrays, each representing the hazard map for a given return period.
                                Each DataArray contains water levels (meters) for the corresponding return period.
        """
        floodmaps = floodmaps.copy()  # avoid modifying the original list
        # Check that all floodmaps have the same shape and dimensions
        first_shape = floodmaps[0].shape
        first_dims = floodmaps[0].dims
        for i, floodmap in enumerate(floodmaps):
            if floodmap.shape != first_shape or floodmap.dims != first_dims:
                raise ValueError(
                    f"Floodmap at index {i} does not match the shape or dimensions of the first floodmap. "
                    f"Expected shape {first_shape} and dims {first_dims}, got shape {floodmap.shape} and dims {floodmap.dims}."
                )

        # Check that zb and mask have the same shape
        if zb.shape != mask.shape:
            raise ValueError(
                "Bed elevation array (zb) and mask must have the same shape."
            )

        # Check that floodmaps, zb, and mask all have the same shape
        if (
            len(first_shape) != len(zb.shape)
            or first_shape != zb.shape
            or first_shape != mask.shape
        ):
            raise ValueError(
                f"Floodmaps, bed elevation array (zb), and mask must all have the same shape. "
                f"Floodmap shape: {first_shape}, zb shape: {zb.shape}, mask shape: {mask.shape}."
            )

        # stack dimensions if floodmaps are 2D
        if len(floodmaps[0].shape) > 1:
            stacking = True
            for i, floodmap in enumerate(floodmaps):
                floodmaps[i] = floodmap.stack(z=("x", "y"))
            zb = zb.stack(z=("x", "y"))
            mask = mask.stack(z=("x", "y"))
        else:
            stacking = False

        # 1a: make a table of all water levels and associated frequencies
        zs = xr.concat(floodmaps, pd.Index(frequencies, name="frequency"))
        # Get the indices of columns with all NaN values
        nan_cells = np.where(np.all(np.isnan(zs), axis=0))[0]
        # fill nan values with minimum bed levels in each grid cell, np.interp cannot ignore nan values
        zs = xr.where(np.isnan(zs), np.tile(zb, (zs.shape[0], 1)), zs)
        # Get table of frequencies
        freq = np.tile(frequencies, (zs.shape[1], 1)).transpose()

        # 1b: sort water levels in descending order and include the frequencies in the sorting process
        # (i.e. each h-value should be linked to the same p-values as in step 1a)
        sort_index = zs.argsort(axis=0)
        sorted_prob = np.flipud(np.take_along_axis(freq, sort_index, axis=0))
        sorted_zs = np.flipud(np.take_along_axis(zs.values, sort_index, axis=0))

        # 1c: Compute exceedance probabilities of water depths
        # Method: accumulate probabilities from top to bottom
        prob_exceed = np.cumsum(sorted_prob, axis=0)

        # 1d: Compute return periods of water depths
        # Method: simply take the inverse of the exceedance probability (1/Pex)
        rp_zs = 1.0 / prob_exceed

        # For each return period (T) of interest do the following:
        # For each grid cell do the following:
        # Use the table from step [1d] as a “lookup-table” to derive the T-year water depth. Use a 1-d interpolation technique:
        # h(T) = interp1 (log(T*), h*, log(T))
        # in which t* and h* are the values from the table and T is the return period (T) of interest
        # The resulting T-year water depths for all grids combined form the T-year hazard map
        rp_da = xr.DataArray(rp_zs, dims=zs.dims)

        # no_data_value = -999  # in SFINCS
        # sorted_zs = xr.where(sorted_zs == no_data_value, np.nan, sorted_zs)

        valid_cells = np.where(mask == 1)[
            0
        ]  # only loop over cells where model is not masked
        h = matlib.repmat(
            np.copy(zb), len(return_periods), 1
        )  # if not flooded (i.e. not in valid_cells) revert to bed_level, read from SFINCS results so it is the minimum bed level in a grid cell

        for jj in valid_cells:  # looping over all non-masked cells.
            # linear interpolation for all return periods to evaluate
            h[:, jj] = np.interp(
                np.log10(return_periods),
                np.log10(rp_da[::-1, jj]),
                sorted_zs[::-1, jj],
                left=0,
            )

        # Re-fill locations that had nan water level for all simulations with nans
        h[:, nan_cells] = np.full(h[:, nan_cells].shape, np.nan)

        # If a cell has the same water-level as the bed elevation it should be dry (turn to nan)
        diff = h - np.tile(zb, (h.shape[0], 1))
        dry = (
            diff < 10e-10
        )  # here we use a small number instead of zero for rounding errors
        h[dry] = np.nan

        rp_maps = []
        for ii, rp in enumerate(return_periods):
            da = xr.DataArray(
                data=h[ii, :], coords={"z": zs["z"]}, attrs={"units": "meters"}
            )
            if stacking:
                # Ensure unstacking creates (y, x) dimensions in the correct order
                da = da.unstack()
                # Reorder dimensions if needed
                if set(da.dims) == {"y", "x"} and da.dims != ("y", "x"):
                    da = da.transpose("y", "x")
            # #create single nc
            rp_maps.append(da)

        return rp_maps



import logging
import time
from functools import wraps

from flood_adapt.misc.log import FloodAdaptLogging


def debug_timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        _logger = FloodAdaptLogging.getLogger()  # No forced log level
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(f"Started '{func.__name__}'")
            start_time = time.perf_counter()

            result = func(*args, **kwargs)

            end_time = time.perf_counter()
            elapsed = end_time - start_time
            _logger.debug(f"Finished '{func.__name__}' in {elapsed:.4f} seconds")
        else:
            result = func(*args, **kwargs)

        return result

    return wrapper



from itertools import combinations

from flood_adapt.dbs_classes.dbs_template import DbsTemplate
from flood_adapt.misc.exceptions import AlreadyExistsError, DatabaseError
from flood_adapt.objects.measures.measures import MeasureType
from flood_adapt.objects.strategies.strategies import Strategy


class DbsStrategy(DbsTemplate[Strategy]):
    dir_name = "strategies"
    display_name = "Strategy"
    _object_class = Strategy
    _higher_lvl_object = "Scenario"

    def get(self, name: str) -> Strategy:
        strategy = super().get(name)
        measures = [
            self._database.measures.get(measure) for measure in strategy.measures
        ]
        strategy.initialize_measure_objects(measures)
        return strategy

    def save(
        self,
        object_model: Strategy,
        overwrite: bool = False,
    ):
        """Save an object in the database and all associated files.

        This saves the toml file and any additional files attached to the object.

        Parameters
        ----------
        object_model : Object
            object to be saved in the database
        overwrite : bool, optional
            whether to overwrite the object if it already exists in the
            database, by default False

        Raises
        ------
        DatabaseError
            Raise error if name is already in use.
        """
        object_exists = object_model.name in self.summarize_objects()["name"]

        # If you want to overwrite the object, and the object already exists, first delete it. If it exists and you
        # don't want to overwrite, raise an error.
        if overwrite and object_exists:
            self.delete(object_model.name, toml_only=True)
        elif not overwrite and object_exists:
            raise AlreadyExistsError(object_model.name, self.display_name)

        # Check if any measures overlap
        self._check_overlapping_measures(object_model.measures)

        # If the folder doesnt exist yet, make the folder and save the object
        if not (self.input_path / object_model.name).exists():
            (self.input_path / object_model.name).mkdir()

        # Save the object and any additional files
        object_model.save(
            self.input_path / object_model.name / f"{object_model.name}.toml",
        )

    def _check_overlapping_measures(self, measures: list[str]):
        """Validate if the combination of impact measures can happen, since impact measures cannot affect the same properties.

        Raises
        ------
        DatabaseError
            information on which combinations of measures have overlapping properties
        """
        measure_objects = [self._database.measures.get(measure) for measure in measures]
        impact_measures = [
            measure
            for measure in measure_objects
            if MeasureType.is_impact(measure.type)
        ]

        adapter = self._database.static.get_fiat_model()

        ids = [adapter.get_object_ids(measure) for measure in impact_measures]

        # Get all possible pairs of measures and check overlapping buildings for each measure
        combs = list(combinations(enumerate(ids), 2))
        common_elements = []
        for comb in combs:
            common_elements.append(list(set(comb[0][1]).intersection(comb[1][1])))
        # If there is any combination with overlapping buildings raise Error and do not allow for Strategy object creation
        overlapping = [len(k) > 0 for k in common_elements]

        if any(overlapping):
            msg = "Cannot create strategy! There are overlapping buildings for which measures are proposed"
            counter = 0
            for i, comb in enumerate(combs):
                if overlapping[i]:
                    if counter > 0:
                        msg += " and"
                    msg += " between '{}' and '{}'".format(
                        impact_measures[comb[0][0]].name,
                        impact_measures[comb[1][0]].name,
                    )
                    counter += 1
            raise DatabaseError(msg)

    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if a strategy is used in a scenario.

        Parameters
        ----------
        name : str
            name of the strategy to be checked

        Returns
        -------
        list[str]
            list of scenarios that use the strategy
        """
        scenarios = [
            self._database.scenarios.get(scn)
            for scn in self._database.scenarios.summarize_objects()["name"]
        ]

        used_in_scenario = [
            scenario.name for scenario in scenarios if name == scenario.strategy
        ]

        return used_in_scenario



import os
from pathlib import Path
from typing import Annotated

import pandas as pd
import xarray as xr

from flood_adapt.misc.utils import (
    copy_file_to_output_dir,
    validate_file_extension,
)
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    IRainfall,
)
from flood_adapt.objects.forcing.netcdf import validate_netcdf_forcing
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.forcing.timeseries import (
    CSVTimeseries,
    SyntheticTimeseries,
    TimeseriesFactory,
)


class RainfallConstant(IRainfall):
    source: ForcingSource = ForcingSource.CONSTANT

    intensity: us.UnitfulIntensity

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        time = pd.date_range(
            start=time_frame.start_time,
            end=time_frame.end_time,
            freq=time_frame.time_step,
            name="time",
        )
        values = [self.intensity.value for _ in range(len(time))]
        return pd.DataFrame(data=values, index=time)


class RainfallSynthetic(IRainfall):
    source: ForcingSource = ForcingSource.SYNTHETIC
    timeseries: SyntheticTimeseries

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        return TimeseriesFactory.from_object(self.timeseries).to_dataframe(
            time_frame=time_frame
        )


class RainfallMeteo(IRainfall):
    source: ForcingSource = ForcingSource.METEO
    precip_units: us.UnitTypesIntensity = us.UnitTypesIntensity.mm_hr
    wind_units: us.UnitTypesVelocity = us.UnitTypesVelocity.mps


class RainfallTrack(IRainfall):
    source: ForcingSource = ForcingSource.TRACK

    path: Annotated[Path, validate_file_extension([".cyc", ".spw"])]

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        if self.path:
            self.path = copy_file_to_output_dir(self.path, Path(output_dir))


class RainfallCSV(IRainfall):
    source: ForcingSource = ForcingSource.CSV

    path: Annotated[Path, validate_file_extension([".csv"])]

    units: us.UnitTypesIntensity = us.UnitTypesIntensity.mm_hr

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        return CSVTimeseries.load_file(
            path=self.path, units=us.UnitfulIntensity(value=0, units=self.units)
        ).to_dataframe(time_frame=time_frame)

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        self.path = copy_file_to_output_dir(self.path, Path(output_dir))


class RainfallNetCDF(IRainfall):
    source: ForcingSource = ForcingSource.NETCDF
    units: us.UnitTypesIntensity = us.UnitTypesIntensity.mm_hr

    path: Annotated[Path, validate_file_extension([".nc"])]

    def read(self) -> xr.Dataset:
        required_vars = ("precip",)
        required_coords = ("time", "lat", "lon")
        with xr.open_dataset(self.path) as ds:
            validated_ds = validate_netcdf_forcing(ds, required_vars, required_coords)
        return validated_ds

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        self.path = copy_file_to_output_dir(self.path, Path(output_dir))



import os
from pathlib import Path
from typing import Literal, Union

import tomli
import tomli_w
from pydantic import BaseModel, Field

from flood_adapt.config.fiat import FiatModel
from flood_adapt.config.gui import GuiModel
from flood_adapt.config.sfincs import SfincsModel


class StandardObjectModel(BaseModel):
    """The accepted input for the variable standard_object in Site."""

    events: list[str] = Field(default_factory=list)
    projections: list[str] = Field(default_factory=list)
    strategies: list[str] = Field(default_factory=list)


class Site(BaseModel):
    """The expected variables and data types of attributes of the Site class.

    Attributes
    ----------
    name : str
        Name of the site.
    description : str
        Description of the site. Defaults to "".
    lat : float
        Latitude of the site.
    lon : float
        Longitude of the site.
    standard_objects : StandardObjectModel, default=StandardObjectModel()
        Standard objects of the site.
    gui : GuiModel
        GUI model of the site.
    sfincs : SfincsModel
        SFincs model of the site.
    fiat : FiatModel
        Fiat model of the site.
    """

    name: str
    description: str = ""
    lat: float
    lon: float
    standard_objects: StandardObjectModel = StandardObjectModel()

    gui: GuiModel
    sfincs: SfincsModel
    fiat: FiatModel

    def save(
        self,
        filepath: Union[str, os.PathLike],
        sfincs: str = "sfincs.toml",
        fiat: str = "fiat.toml",
        gui: str = "gui.toml",
    ) -> None:
        """Write toml file from model object."""
        parent_folder = Path(filepath).parent
        config_dict = {
            "name": self.name,
            "description": self.description,
            "lat": self.lat,
            "lon": self.lon,
            "standard_objects": {
                "events": self.standard_objects.events,
                "projections": self.standard_objects.projections,
                "strategies": self.standard_objects.strategies,
            },
            "components": {},
        }

        if self.sfincs is not None:
            config_dict["components"]["sfincs"] = sfincs
            with open(parent_folder / sfincs, "wb") as f:
                tomli_w.dump(self.sfincs.model_dump(exclude_none=True), f)

        if self.fiat is not None:
            config_dict["components"]["fiat"] = fiat
            with open(parent_folder / fiat, "wb") as f:
                tomli_w.dump(self.fiat.model_dump(exclude_none=True), f)

        if self.gui is not None:
            config_dict["components"]["gui"] = gui
            with open(parent_folder / gui, "wb") as f:
                tomli_w.dump(self.gui.model_dump(exclude_none=True), f)

        with open(filepath, "wb") as f:
            tomli_w.dump(config_dict, f)

    @staticmethod
    def load_file(filepath: Union[str, os.PathLike]) -> "Site":
        """Create Site from toml file."""
        return SiteBuilder.load_file(Path(filepath))


class SiteBuilder(BaseModel):
    """Pydantic model that reads the site configuration file and builds the site model.

    Note that the components are not required, as the site may not have all of them.
    Second, note that the components are assumed to be the file names of the component configs, located in the same directory as the site configuration file.
    """

    name: str
    description: str = ""
    lat: float
    lon: float
    components: dict[Literal["sfincs", "fiat", "gui"], str]

    @staticmethod
    def load_file(file_path: Path) -> "Site":
        with open(file_path, "rb") as f:
            model_dict = tomli.load(f)

        toml_dir = file_path.parent
        if (sfincs_config := model_dict["components"].get("sfincs")) is not None:
            model_dict["sfincs"] = SfincsModel.read_toml(toml_dir / sfincs_config)

        if (gui_config := model_dict["components"].get("gui")) is not None:
            model_dict["gui"] = GuiModel.read_toml(toml_dir / gui_config)

        if (fiat_config := model_dict["components"].get("fiat")) is not None:
            model_dict["fiat"] = FiatModel.read_toml(toml_dir / fiat_config)

        return Site(**model_dict)



from abc import abstractmethod
from typing import Any

import geopandas as gpd

from flood_adapt.adapter.interface.model_adapter import IAdapter
from flood_adapt.objects.forcing.forcing import IForcing
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.measures.measures import Measure
from flood_adapt.objects.projections.projections import Projection


class IHazardAdapter(IAdapter):
    @abstractmethod
    def set_timing(self, time: TimeFrame):
        """
        Implement this to handle the timing of the event from the Event.

        Access the events timing by `event.timing`, which contains the start and end time of the event, and the time step.
        """
        pass

    @abstractmethod
    def add_forcing(self, forcing: IForcing):
        """
        Implement this to handle each supported forcing type for this Hazard model.

        A forcing is time series data, i.e. water levels, wind, rain, discharge, with optional time and space dimensions.
        So this could be a single time series, or a collection of time series, or a time series associated with a spatial grid/gridpoint)

        Forcings contain all information needed to implement the forcing in the hazard model. (geospatial files, parameters, etc.)
        """
        pass

    @abstractmethod
    def add_measure(self, measure: Measure):
        """
        Implement this to handle each supported measure type for this Hazard model.

        A hazardmeasure is a measure that affects the hazard model, i.e. a measure that affects the water levels, wind, rain, discharge.
        For example a measure could be a dike, a land use change, a change in the river channel, etc.

        HazardMeasures contain all information needed to implement the measure in the hazard model. (geospatial files, parameters, etc.)

        """
        pass

    @abstractmethod
    def add_projection(self, projection: Projection):
        """
        Implement this to handle each supported projection type for this Hazard model.

        A projection is a projection of the future, i.e. sea level rise, subsidence, rainfall multiplier, storm frequency increase, etc.
        PhysicalProjections contains all information needed to implement the projection in the hazard model. (parameters, etc.)
        """
        pass

    @abstractmethod
    def get_model_boundary(self) -> gpd.GeoDataFrame:
        """
        Implement this to return the model boundary of the hazard model.

        The model boundary is a geospatial file that defines the boundary of the hazard model.
        """
        pass

    @abstractmethod
    def get_model_grid(self) -> Any:
        """Implement this to return the model grid of the hazard model."""
        pass



from os import PathLike
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

import tomli_w
from pydantic import BaseModel, Field, field_validator, model_validator

from flood_adapt.adapter.fiat_adapter import _IMPACT_COLUMNS


def combine_filters(*filters):
    """
    Combine multiple SQL filter strings with AND operators.

    Parameters
    ----------
    *filters : str
        Variable number of filter strings to combine.

    Returns
    -------
    str
        Combined filter string with AND operators, excluding empty filters.
    """
    return " AND ".join(f for f in filters if f)


def pascal_case(s):
    """
    Convert a string to PascalCase.

    Parameters
    ----------
    s : str
        Input string to convert.

    Returns
    -------
    str
        String converted to PascalCase.
    """
    return "".join(word.capitalize() for word in s.split())


class FieldMapping(BaseModel):
    """
    Represents a mapping of a database field to a list of allowed values.

    Parameters
    ----------
    field_name : str
        The name of the database field/column
    values : List[str]
        List of values that should match this field

    Methods
    -------
    to_sql_filter()
        Generate SQL filter string for this field mapping.
    """

    field_name: str
    values: List[str]

    def to_sql_filter(self) -> str:
        """
        Generate SQL filter string for this field mapping.

        Returns
        -------
        str
            SQL WHERE clause condition string for this field mapping.
        """
        quoted_values = ", ".join([f"'{v}'" for v in self.values])
        return f"`{self.field_name}` IN ({quoted_values})"


class TypeMapping(BaseModel):
    """
    Container for multiple field mappings that define object type filtering.

    Parameters
    ----------
    mappings : List[FieldMapping]
        List of field mappings that together define the type criteria

    Methods
    -------
    add_mapping(field_name, values)
        Add a new field mapping.
    to_sql_filter()
        Generate combined SQL filter string from all mappings.
    """

    mappings: List[FieldMapping] = Field(default_factory=list)

    def add_mapping(self, field_name: str, values: List[str]) -> None:
        """
        Add a new field mapping.

        Parameters
        ----------
        field_name : str
            Name of the database field.
        values : List[str]
            List of allowed values for this field.
        """
        self.mappings.append(FieldMapping(field_name=field_name, values=values))

    def to_sql_filter(self) -> str:
        """
        Generate combined SQL filter string from all mappings.

        Returns
        -------
        str
            Combined SQL WHERE clause condition string.
        """
        if not self.mappings:
            return ""
        filter_parts = [mapping.to_sql_filter() for mapping in self.mappings]
        return " AND ".join(filter_parts)


class MetricModel(BaseModel):
    """
    Represents a metric configuration for infometric analysis.

    Parameters
    ----------
    name : str
        The short name of the metric.
    long_name : Optional[str], default=None
        The long descriptive name of the metric. If not provided, defaults to `name`.
    show_in_metrics_table : Optional[bool], default=True
        Indicates whether the metric should be displayed in the metrics table.
    show_in_map : Optional[bool], default=True
        Indicates whether the metric should be displayed on the map.
    description : Optional[str], default=None
        A detailed description of the metric. If not provided, defaults to `name`.
    select : str
        The SQL select statement or expression for the metric.
    filter : Optional[str], default=""
        An optional SQL filter to apply to the metric. Defaults to no filter.

    Validation
    ----------
    If `long_name` or `description` are None, they will be set to the value of `name`.
    """

    name: str
    long_name: Optional[str] = None
    show_in_metrics_table: Optional[bool] = True
    show_in_map: Optional[bool] = True
    description: Optional[str] = None
    select: str
    filter: Optional[str] = ""  # This defaults to no filter

    @model_validator(mode="after")
    def fill_defaults(self):
        # If long_name is missing, copy name
        if self.long_name is None:
            self.long_name = self.name

        # If description is missing, copy long_name (which is now guaranteed)
        if self.description is None:
            self.description = self.long_name

        return self


class ImpactCategoriesModel(BaseModel):
    """
    Model for defining impact categories with associated colors, field, unit, and bins.

    Parameters
    ----------
    categories : list[str], default=["Minor", "Major", "Severe"]
        List of impact category names.
    colors : Optional[list[str]], default=["#ffa500", "#ff0000", "#000000"]
        List of colors corresponding to each category.
    field : str
        The database field name used for categorization.
    unit : str
        The unit of measurement for the field.
    bins : list[float]
        List of threshold values for binning the field values.

    Methods
    -------
    validate_colors_length(colors, info)
        Validate that colors list length matches categories list length.
    validate_bins_length(bins, info)
        Validate that bins list length is one less than categories list length.
    """

    categories: list[str] = Field(default_factory=lambda: ["Minor", "Major", "Severe"])
    colors: Optional[list[str]] = Field(
        default_factory=lambda: ["#ffa500", "#ff0000", "#000000"]
    )
    field: str = _IMPACT_COLUMNS.inundation_depth
    unit: str
    bins: list[float]

    @field_validator("colors", mode="before")
    @classmethod
    def validate_colors_length(cls, colors, info):
        """
        Validate that colors list length matches categories list length.

        Parameters
        ----------
        colors : list[str]
            List of color values.
        info : Any
            Field validation info containing all field values.

        Returns
        -------
        list[str]
            The validated colors list.

        Raises
        ------
        ValueError
            If colors length doesn't match categories length.
        """
        categories = info.data.get("categories")
        if categories and colors and len(colors) != len(categories):
            raise ValueError("Length of 'colors' must match length of 'categories'.")
        return colors

    @field_validator("bins", mode="before")
    @classmethod
    def validate_bins_length(cls, bins, info):
        """
        Validate that bins list length is one less than categories list length.

        Parameters
        ----------
        bins : list[float]
            List of bin threshold values.
        info : Any
            Field validation info containing all field values.

        Returns
        -------
        list[float]
            The validated bins list.

        Raises
        ------
        ValueError
            If bins length is not one less than categories length.
        """
        categories = info.data.get("categories")
        if categories and len(bins) != len(categories) - 1:
            raise ValueError(
                "Length of 'bins' must be one less than length of 'categories'."
            )
        return bins


class BuildingsInfographicModel(BaseModel):
    """
    Model for building infographic configuration.

    Parameters
    ----------
    types : list[str]
        List of building types.
    icons : list[str]
        List of icon names corresponding to each building type.
    type_mapping : dict[str, TypeMapping]
        Mapping of building types to their database filtering criteria.
    impact_categories : ImpactCategoriesModel
        Impact categories configuration.

    Methods
    -------
    validate_icons_length(icons, info)
        Validate that icons list length matches types list length.
    get_template(type)
        Get a pre-configured template for OSM or NSI building types.
    """

    # Define building types
    types: list[str]
    icons: list[str]
    type_mapping: dict[str, TypeMapping]
    # Define impact categories
    impact_categories: ImpactCategoriesModel

    @field_validator("icons", mode="before")
    @classmethod
    def validate_icons_length(cls, icons, info):
        """
        Validate that icons list length matches types list length.

        Parameters
        ----------
        icons : list[str]
            List of icon names.
        info : Any
            Field validation info containing all field values.

        Returns
        -------
        list[str]
            The validated icons list.

        Raises
        ------
        ValueError
            If icons length doesn't match types length.
        """
        types = info.data.get("types")
        if types and len(icons) != len(types):
            raise ValueError("Length of 'icons' must equal to the length of 'types'.")
        return icons

    @staticmethod
    def get_template(type: Literal["OSM", "NSI"]):
        """
        Get a pre-configured template for building infographics.

        Parameters
        ----------
        type : Literal["OSM", "NSI"]
            The database type to create a template for.

        Returns
        -------
        BuildingsInfographicModel
            Pre-configured building infographic model.
        """
        if type == "OSM":
            config = BuildingsInfographicModel(
                types=["Residential", "Commercial", "Industrial"],
                icons=["house", "cart", "factory"],
                type_mapping={
                    "Residential": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name=_IMPACT_COLUMNS.primary_object_type,
                                values=["residential"],
                            )
                        ]
                    ),
                    "Commercial": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name=_IMPACT_COLUMNS.primary_object_type,
                                values=["commercial"],
                            )
                        ]
                    ),
                    "Industrial": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name=_IMPACT_COLUMNS.primary_object_type,
                                values=["industrial"],
                            )
                        ]
                    ),
                },
                impact_categories=ImpactCategoriesModel(
                    unit="meters", bins=[0.25, 1.5]
                ),
            )
        elif type == "NSI":
            config = BuildingsInfographicModel(
                types=[
                    "Residential",
                    "Commercial",
                    "Health facilities",
                    "Schools",
                    "Emergency facilities",
                ],
                icons=["house", "cart", "hospital", "school", "firetruck"],
                type_mapping={
                    "Residential": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name=_IMPACT_COLUMNS.primary_object_type,
                                values=["RES"],
                            )
                        ]
                    ),
                    "Commercial": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name="Secondary Object Type",
                                values=[
                                    "COM1",
                                    "COM2",
                                    "COM3",
                                    "COM4",
                                    "COM5",
                                    "COM8",
                                    "COM9",
                                ],
                            )
                        ]
                    ),
                    "Health facilities": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name="Secondary Object Type",
                                values=["RES6", "COM6", "COM7"],
                            )
                        ]
                    ),
                    "Schools": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name="Secondary Object Type",
                                values=["EDU1", "EDU2"],
                            )
                        ]
                    ),
                    "Emergency facilities": TypeMapping(
                        mappings=[
                            FieldMapping(
                                field_name="Secondary Object Type", values=["GOV2"]
                            )
                        ]
                    ),
                },
                impact_categories=ImpactCategoriesModel(unit="feet", bins=[1, 6]),
            )
        return config


class SviModel(BaseModel):
    """
    Model for Social Vulnerability Index (SVI) configuration.

    Parameters
    ----------
    classes : list[str], default=["Low", "High"]
        List of vulnerability class names.
    colors : list[str], default=["#D5DEE1", "#88A2AA"]
        List of colors corresponding to each vulnerability class.
    thresholds : list[float], default=[0.7]
        List of threshold values for vulnerability classification.

    Methods
    -------
    validate_colors_length(colors, info)
        Validate that colors list length matches classes list length.
    validate_thresholds_length(thresholds, info)
        Validate that thresholds list length is one less than classes list length.
    """

    classes: list[str] = Field(default_factory=lambda: ["Low", "High"])
    colors: list[str] = Field(default_factory=lambda: ["#D5DEE1", "#88A2AA"])
    thresholds: list[float] = Field(default_factory=lambda: [0.7])

    @field_validator("colors", mode="before")
    @classmethod
    def validate_colors_length(cls, colors, info):
        """
        Validate that colors list length matches classes list length.

        Parameters
        ----------
        colors : list[str]
            List of color values.
        info : Any
            Field validation info containing all field values.

        Returns
        -------
        list[str]
            The validated colors list.

        Raises
        ------
        ValueError
            If colors length doesn't match classes length.
        """
        classes = info.data.get("classes")
        if classes and colors and len(colors) != len(classes):
            raise ValueError("Length of 'colors' must match length of 'classes'.")
        return colors

    @field_validator("thresholds", mode="before")
    @classmethod
    def validate_thresholds_length(cls, thresholds, info):
        """
        Validate that thresholds list length is one less than classes list length.

        Parameters
        ----------
        thresholds : list[float]
            List of threshold values.
        info : Any
            Field validation info containing all field values.

        Returns
        -------
        list[float]
            The validated thresholds list.

        Raises
        ------
        ValueError
            If thresholds length is not one less than classes length.
        """
        classes = info.data.get("classes")
        if classes and len(thresholds) != len(classes) - 1:
            raise ValueError(
                "Length of 'thresholds' must be one less than length of 'classes'."
            )
        return thresholds


class HomesInfographicModel(BaseModel):
    """
    Model for Homes and SVI (Social Vulnerability Index) infographic configuration.

    Parameters
    ----------
    svi : SviModel
        SVI classification configuration.
    mapping : TypeMapping
        Database field mapping for filtering relevant objects.
    impact_categories : ImpactCategoriesModel
        Impact categories configuration.

    Methods
    -------
    get_template(svi_threshold, type)
        Get a pre-configured template for SVI infographics.
    """

    svi: Optional[SviModel] = None
    mapping: TypeMapping
    impact_categories: ImpactCategoriesModel

    @staticmethod
    def get_template(
        type: Literal["OSM", "NSI"] = "OSM", svi_threshold: Optional[float] = None
    ):
        """
        Get a pre-configured template for SVI infographics.

        Parameters
        ----------
        svi_threshold : Optional[float], default=None
            The SVI threshold value for vulnerability classification. If not provided, SVI will be None.
        type : Literal["OSM", "NSI"], default="OSM"
            The database type to create a template for.

        Returns
        -------
        HomesInfographicModel
            Pre-configured Homes infographic model.
        """
        if svi_threshold is not None:
            svi_model = SviModel(thresholds=[svi_threshold])
        else:
            svi_model = None

        if type == "OSM":
            config = HomesInfographicModel(
                svi=svi_model,
                mapping=TypeMapping(
                    mappings=[
                        FieldMapping(
                            field_name=_IMPACT_COLUMNS.primary_object_type,
                            values=["residential"],
                        )
                    ]
                ),
                impact_categories=ImpactCategoriesModel(
                    categories=["Flooded", "Displaced"],
                    colors=None,
                    unit="meters",
                    bins=[1.5],
                ),
            )
        elif type == "NSI":
            config = HomesInfographicModel(
                svi=svi_model,
                mapping=TypeMapping(
                    mappings=[
                        FieldMapping(
                            field_name=_IMPACT_COLUMNS.primary_object_type,
                            values=["RES"],
                        )
                    ]
                ),
                impact_categories=ImpactCategoriesModel(
                    categories=["Flooded", "Displaced"],
                    colors=None,
                    unit="feet",
                    bins=[6],
                ),
            )
        return config


class RoadsInfographicModel(BaseModel):
    """
    Model for roads infographic configuration.

    Parameters
    ----------
    categories : list[str], default=["Slight", "Minor", "Major", "Severe"]
        List of road impact category names.
    colors : list[str], default=["#e0f7fa", "#80deea", "#26c6da", "#006064"]
        List of colors corresponding to each category.
    icons : list[str], default=["walking_person", "car", "truck", "ambulance"]
        List of icon names for each category.
    users : list[str], default=["Pedestrians", "Cars", "Trucks", "Rescue vehicles"]
        List of road user types for each category.
    thresholds : list[float]
        List of threshold values for categorizing road impacts.
    field : str, default=_IMPACT_COLUMNS.inundation_depth
        The database field name used for categorization.
    unit : str
        The unit of measurement for the field.
    road_length_field : str, default=_IMPACT_COLUMNS.segment_length
        The database field name containing road segment lengths.

    Methods
    -------
    validate_lengths(v, info)
        Validate that all list attributes have the same length.
    get_template(unit_system)
        Get a pre-configured template for metric or imperial units.
    """

    categories: list[str] = Field(
        default_factory=lambda: ["Slight", "Minor", "Major", "Severe"]
    )
    colors: list[str] = Field(
        default_factory=lambda: ["#D5DEE1", "#D5DEE1", "#D5DEE1", "#D5DEE1"]
    )
    icons: list[str] = Field(
        default_factory=lambda: ["walking_person", "car", "truck", "ambulance"]
    )
    users: list[str] = Field(
        default_factory=lambda: ["Pedestrians", "Cars", "Trucks", "Rescue vehicles"]
    )
    thresholds: list[float]
    field: str = _IMPACT_COLUMNS.inundation_depth
    unit: str
    road_length_field: str = _IMPACT_COLUMNS.segment_length

    @field_validator("categories", mode="after")
    @classmethod
    def validate_lengths(cls, v, info):
        """
        Validate that all list attributes have the same length.

        Parameters
        ----------
        v : list[str]
            The categories list.
        info : Any
            Field validation info containing all field values.

        Returns
        -------
        list[str]
            The validated categories list.

        Raises
        ------
        ValueError
            If list attributes don't have the same length.
        """
        # Check that categories, colors, icons, users, thresholds have the same length
        attrs = ["categories", "colors", "icons", "users", "thresholds"]
        lengths = [len(info.data.get(attr, [])) for attr in attrs]
        if len(set(lengths)) > 1:
            raise ValueError(
                f"Attributes {attrs} must all have the same length, got lengths: {lengths}"
            )
        return v

    @staticmethod
    def get_template(unit_system: Literal["metric", "imperial"]):
        """
        Get a pre-configured template for roads infographics.

        Parameters
        ----------
        unit_system : Literal["metric", "imperial"]
            The unit system to use for thresholds and measurements.

        Returns
        -------
        RoadsInfographicModel
            Pre-configured roads infographic model.
        """
        if unit_system == "metric":
            config = RoadsInfographicModel(
                thresholds=[0.1, 0.2, 0.4, 0.8],
                unit="meters",
            )
        elif unit_system == "imperial":
            config = RoadsInfographicModel(
                thresholds=[0.3, 0.5, 1, 2],
                unit="feet",
            )
        return config


class EventInfographicModel(BaseModel):
    """
    Model for event-based infographic configuration.

    Parameters
    ----------
    buildings : Optional[BuildingsInfographicModel], default=None
        Buildings infographic configuration.
    svi : Optional[SviInfographicModel], default=None
        SVI infographic configuration.
    roads : Optional[RoadsInfographicModel], default=None
        Roads infographic configuration.
    """

    buildings: Optional[BuildingsInfographicModel] = None
    svi: Optional[HomesInfographicModel] = None
    roads: Optional[RoadsInfographicModel] = None


class FloodExceedanceModel(BaseModel):
    """
    Model for flood exceedance probability configuration.

    Parameters
    ----------
    column : str, default=_IMPACT_COLUMNS.inundation_depth
        The database column name for flood depth measurements.
    threshold : float, default=0.1
        The flood depth threshold value.
    unit : str, default="meters"
        The unit of measurement for the threshold.
    period : int, default=30
        The time period in years for exceedance analysis.
    """

    column: str = _IMPACT_COLUMNS.inundation_depth
    threshold: float = 0.1
    unit: str = "meters"
    period: int = 30


class RiskInfographicModel(BaseModel):
    """
    Model for risk-based infographic configuration.

    Parameters
    ----------
    homes : HomesInfographicModel
        Homes infographic configuration.
    flood_exceedances : FloodExceedanceModel
        Flood exceedance configuration.

    Methods
    -------
    get_template(type, svi_threshold)
        Get a pre-configured template for risk infographics.
    """

    homes: HomesInfographicModel
    flood_exceedance: FloodExceedanceModel

    @staticmethod
    def get_template(
        type: Literal["OSM", "NSI"], svi_threshold: Optional[float] = None
    ):
        """
        Get a pre-configured template for risk infographics.

        Parameters
        ----------
        type : Literal["OSM", "NSI"]
            The database type to create a template for.
        svi_threshold : Optional[float], default=None
            The SVI threshold value for vulnerability classification.

        Returns
        -------
        RiskInfographicModel
            Pre-configured risk infographic model.
        """
        if type == "OSM":
            config = RiskInfographicModel(
                homes=HomesInfographicModel.get_template(
                    type="OSM", svi_threshold=svi_threshold
                ),
                flood_exceedance=FloodExceedanceModel(),
            )
        elif type == "NSI":
            config = RiskInfographicModel(
                homes=HomesInfographicModel.get_template(
                    type="NSI", svi_threshold=svi_threshold
                ),
                flood_exceedance=FloodExceedanceModel(unit="feet", threshold=0.2),
            )
        return config


def get_filter(
    type_mapping: TypeMapping,
    cat_field: str,
    cat_idx: int,
    bins: list[float],
    base_filt="",
) -> str:
    """
    Construct a SQL filter string based on provided type mapping and category criteria.

    Parameters
    ----------
    type_mapping : TypeMapping
        TypeMapping object containing field mappings to filter on.
    cat_field : str
        Name of the field representing the category in the database.
    cat_idx : int
        Index indicating which category bin to use for filtering.
    bins : list[float]
        List of bin thresholds for the category field.
    base_filt : str, default=""
        Additional base filter string to prepend.

    Returns
    -------
    str
        A SQL filter string combining type and category conditions.
    """
    # Build type filters using TypeMapping
    type_filter = type_mapping.to_sql_filter()

    # Add category filter
    if cat_idx == 0:
        cat_filter = f"`{cat_field}` <= {bins[0]}"
    elif cat_idx == len(bins):
        cat_filter = f"`{cat_field}` > {bins[-1]}"
    else:
        cat_filter = (
            f"`{cat_field}` <= {bins[cat_idx]} AND `{cat_field}` > {bins[cat_idx-1]}"
        )

    return combine_filters(base_filt, type_filter, cat_filter)


class Metrics:
    """Main class for managing impact metrics configuration and generation."""

    def __init__(self, dmg_unit: str, return_periods: list[float]):
        """
        Initialize the Metrics class.

        Parameters
        ----------
        dmg_unit : str
            The unit of measurement for damage values.
        return_periods : list[float]
            List of return periods in years for risk analysis.
        """
        self.dmg_unit = dmg_unit
        self.return_periods = return_periods

        # Initialize all metric lists as empty instance attributes
        self.mandatory_metrics_event: list[MetricModel] = []
        self.mandatory_metrics_risk: list[MetricModel] = []
        self.additional_metrics_event: list[MetricModel] = []
        self.additional_metrics_risk: list[MetricModel] = []
        self.infographics_metrics_event: list[MetricModel] = []
        self.infographics_metrics_risk: list[MetricModel] = []
        self.additional_risk_configs: dict = {}
        self.infographics_config: dict = {}

    @staticmethod
    def write_metrics(metrics, path, aggr_levels=[]):
        """
        Write metrics configuration to a TOML file.

        Parameters
        ----------
        metrics : list[MetricModel]
            List of metric models to write.
        path : Union[str, Path]
            Path to the output TOML file.
        aggr_levels : list[str], default=[]
            List of aggregation levels.
        """
        attrs = {}
        attrs["aggregateBy"] = aggr_levels
        attrs["queries"] = [metric.model_dump() for metric in metrics]

        # Save metrics configuration
        with open(path, "wb") as f:
            tomli_w.dump(attrs, f)

    def write(
        self,
        metrics_path: Union[str, Path, PathLike],
        aggregation_levels: List[str],
        infographics_path: Optional[Union[str, Path, PathLike]] = None,
    ) -> None:
        """
        Write all metrics (mandatory, additional, and infographics) to TOML files.

        Parameters
        ----------
        metrics_path : Union[str, Path, PathLike]
            The directory path where the metrics configuration files will be saved.
        aggregation_levels : List[str]
            A list of aggregation levels to include in the metrics configuration files.
        infographics_path : Optional[Union[str, Path, PathLike]], default=None
            The directory path where infographics configuration files will be saved.
            Required if infographics configurations are present.

        Raises
        ------
        ValueError
            If infographics_path is None but infographics configurations exist.
        """
        path_im = Path(metrics_path)
        path_im.mkdir(parents=True, exist_ok=True)

        # Write mandatory event metrics
        self.write_metrics(
            self.mandatory_metrics_event,
            path_im / "mandatory_metrics_config.toml",
            aggregation_levels,
        )

        # Write mandatory risk metrics
        if self.mandatory_metrics_risk:
            self.write_metrics(
                self.mandatory_metrics_risk,
                path_im / "mandatory_metrics_config_risk.toml",
                aggregation_levels,
            )

        # Write additional event metrics if any
        if self.additional_metrics_event:
            self.write_metrics(
                self.additional_metrics_event,
                path_im / "additional_metrics_config.toml",
                aggregation_levels,
            )

        # Write additional risk metrics if any
        if self.additional_metrics_risk:
            self.write_metrics(
                self.additional_metrics_risk,
                path_im / "additional_metrics_config_risk.toml",
                aggregation_levels,
            )

        # Write infographics event metrics if any
        if (
            hasattr(self, "infographics_metrics_event")
            and self.infographics_metrics_event
        ):
            self.write_metrics(
                self.infographics_metrics_event,
                path_im / "infographic_metrics_config.toml",
                aggregation_levels,
            )

        # Write infographics risk metrics if any
        if (
            hasattr(self, "infographics_metrics_risk")
            and self.infographics_metrics_risk
        ):
            self.write_metrics(
                self.infographics_metrics_risk,
                path_im / "infographic_metrics_config_risk.toml",
                aggregation_levels,
            )

        # Save additional risk configurations if available
        if hasattr(self, "additional_risk_configs") and self.additional_risk_configs:
            with open(path_im / "metrics_additional_risk_configs.toml", "wb") as f:
                tomli_w.dump(self.additional_risk_configs, f)

        # Save infographics configuration if available
        if self.infographics_config:
            if infographics_path is None:
                raise ValueError(
                    "infographics_path must be provided to save infographics configuration."
                )
            infographics_path = Path(infographics_path)
            infographics_path.mkdir(parents=True, exist_ok=True)
            if "buildings" in self.infographics_config:
                with open(infographics_path / "config_charts.toml", "wb") as f:
                    tomli_w.dump(self.infographics_config["buildings"], f)
            if "svi" in self.infographics_config:
                with open(infographics_path / "config_people.toml", "wb") as f:
                    tomli_w.dump(self.infographics_config["svi"], f)
            if "roads" in self.infographics_config:
                with open(infographics_path / "config_roads.toml", "wb") as f:
                    tomli_w.dump(self.infographics_config["roads"], f)
            if "risk" in self.infographics_config:
                with open(infographics_path / "config_risk_charts.toml", "wb") as f:
                    tomli_w.dump(self.infographics_config["risk"], f)

    def create_mandatory_metrics_event(self) -> list[MetricModel]:
        """
        Create mandatory metrics for event analysis.

        Returns
        -------
        list[MetricModel]
            List of mandatory event metrics.
        """
        self.mandatory_metrics_event.append(
            MetricModel(
                name="TotalDamageEvent",
                description="Total building damage",
                long_name=f"Total building damage ({self.dmg_unit})",
                select=f"SUM(`{_IMPACT_COLUMNS.total_damage}`)",
                filter="",
                show_in_metrics_table=True,
            )
        )
        return self.mandatory_metrics_event

    def create_mandatory_metrics_risk(self) -> list[MetricModel]:
        """
        Create mandatory metrics for risk analysis.

        Returns
        -------
        list[MetricModel]
            List of mandatory risk metrics.
        """
        self.mandatory_metrics_risk.append(
            MetricModel(
                name="ExpectedAnnualDamages",
                description="Expected annual damages",
                long_name=f"Expected annual damages ({self.dmg_unit})",
                select=f"SUM(`{_IMPACT_COLUMNS.risk_ead}`)",
                filter="",
                show_in_metrics_table=True,
            )
        )
        for rp in self.return_periods:
            self.mandatory_metrics_risk.append(
                MetricModel(
                    name=f"TotalDamageRP{int(rp)}",
                    description=f"Total damage with return period of {int(rp)} years",
                    long_name=f"Total building damage - {int(rp)}Y ({self.dmg_unit})",
                    select=f"SUM(`{_IMPACT_COLUMNS.total_damage_rp.format(years=int(rp))}`)",
                    filter="",
                    show_in_metrics_table=True,
                )
            )

        return self.mandatory_metrics_risk

    def add_event_metric(self, metric: MetricModel) -> None:
        """
        Add an additional event metric.

        Parameters
        ----------
        metric : MetricModel
            The metric to add to the additional event metrics list.

        Raises
        ------
        ValueError
            If a metric with the same name already exists.
        """
        if any(m.name == metric.name for m in self.additional_metrics_event):
            raise ValueError(f"Event metric with name '{metric.name}' already exists.")
        self.additional_metrics_event.append(metric)

    def add_risk_metric(self, metric: MetricModel) -> None:
        """
        Add an additional risk metric.

        Parameters
        ----------
        metric : MetricModel
            The metric to add to the additional risk metrics list.

        Raises
        ------
        ValueError
            If a metric with the same name already exists.
        """
        if any(m.name == metric.name for m in self.additional_metrics_risk):
            raise ValueError(f"Risk metric with name '{metric.name}' already exists.")
        self.additional_metrics_risk.append(metric)

    def create_infographics_metrics_event(
        self,
        config: EventInfographicModel,
        base_filt=f"`{_IMPACT_COLUMNS.total_damage}` > 0",
    ) -> list[MetricModel]:
        """
        Create infographic metrics for event analysis.

        Parameters
        ----------
        config : EventInfographicModel
            Configuration for event infographics.
        base_filt : str, default="`Total Damage` > 0"
            Base SQL filter to apply to all metrics.

        Returns
        -------
        list[MetricModel]
            List of infographic event metrics.
        """
        # Generate queries for all building types and categories
        if config.buildings:
            self._setup_buildings(config.buildings, base_filt)
        if config.svi:
            self._setup_svi(config.svi, base_filt)
        if config.roads:
            self._setup_roads(config.roads)
        return self.infographics_metrics_event

    def create_infographics_metrics_risk(
        self,
        config: RiskInfographicModel,
        base_filt=f"`{_IMPACT_COLUMNS.risk_ead}` > 0",
    ) -> list[MetricModel]:
        """
        Create infographic metrics for risk analysis.

        Parameters
        ----------
        config : RiskInfographicModel
            Configuration for risk infographics.
        base_filt : str, default="`Risk (EAD)` > 0"
            Base SQL filter to apply to all metrics.

        Returns
        -------
        list[MetricModel]
            List of infographic risk metrics.
        """
        infographics_metrics_risk = []

        # Get mapping from config.svi
        mapping = config.homes.mapping

        # Build type filter string using TypeMapping
        type_cond = mapping.to_sql_filter()

        # FloodedHomes (Exceedance Probability > 50)
        fe = config.flood_exceedance
        filter_str = combine_filters(
            "`Exceedance Probability` > 50", type_cond, base_filt
        )
        infographics_metrics_risk.append(
            MetricModel(
                name="LikelyFloodedHomes",
                description=f"Homes likely to flood ({fe.column} > {fe.threshold} {fe.unit}) in {fe.period} year period",
                select="COUNT(*)",
                filter=filter_str,
                long_name=f"Homes likely to flood in {fe.period}-year period (#)",
                show_in_metrics_table=True,
            )
        )

        # ImpactedHomes for each RP - use class return periods and config SVI thresholds
        rps = self.return_periods
        if config.homes.svi is not None:
            svi_thresholds = config.homes.svi.thresholds
            svi_classes = config.homes.svi.classes
        else:
            svi_thresholds = []
            svi_classes = []

        for rp in rps:
            # ImpactedHomes{RP} (all homes)
            filter_str = combine_filters(
                f"`{fe.column} ({int(rp)}Y)` >= {fe.threshold}", type_cond, base_filt
            )
            infographics_metrics_risk.append(
                MetricModel(
                    name=f"ImpactedHomes{int(rp)}Y",
                    description=f"Number of homes impacted ({fe.column} > {fe.threshold} {fe.unit}) in the {int(rp)}-year event",
                    select="COUNT(*)",
                    filter=filter_str,
                    long_name=f"Flooded homes - RP{int(rp)} (#)",
                    show_in_metrics_table=True,
                )
            )

            # Create metrics for each SVI class
            for j, svi_class in enumerate(svi_classes):
                # Build SVI condition based on thresholds
                if j == 0:
                    # First class: SVI < first_threshold
                    svi_cond = f"`SVI` < {svi_thresholds[0]}"
                elif j == len(svi_classes) - 1:
                    # Last class: SVI >= last_threshold
                    svi_cond = f"`SVI` >= {svi_thresholds[-1]}"
                else:
                    # Middle classes: previous_threshold <= SVI < current_threshold
                    svi_cond = f"`SVI` >= {svi_thresholds[j-1]} AND `SVI` < {svi_thresholds[j]}"

                # Clean class name for metric naming (remove spaces, special chars)
                clean_class_name = svi_class.replace(" ", "").replace("-", "")

                filter_str = combine_filters(
                    f"`{fe.column} ({int(rp)}Y)` >= {fe.threshold}",
                    type_cond,
                    svi_cond,
                    base_filt,
                )

                infographics_metrics_risk.append(
                    MetricModel(
                        name=f"ImpactedHomes{int(rp)}Y{clean_class_name}SVI",
                        description=f"{svi_class} vulnerable homes impacted ({fe.column} > {fe.threshold} {fe.unit}) in the {int(rp)}-year event",
                        select="COUNT(*)",
                        filter=filter_str,
                        long_name=f"Flooded homes with {svi_class} vulnerability - RP{int(rp)} (#)",
                        show_in_metrics_table=True,
                    )
                )

        self.infographics_metrics_risk = infographics_metrics_risk
        self.additional_risk_configs = {
            "flood_exceedance": {
                "column": fe.column,
                "threshold": fe.threshold,
                "period": fe.period,
            }
        }
        self._make_infographics_config_risk(config)
        return self.infographics_metrics_risk

    def _setup_buildings(self, config: BuildingsInfographicModel, base_filt) -> None:
        """
        Configure building metrics and configuration for infographics.

        Parameters
        ----------
        config : BuildingsInfographicModel
            Building infographic configuration.
        base_filt : str
            Base SQL filter to apply.
        """
        # Generate queries for all building types and categories
        building_queries = []
        for btype in config.types:
            type_mapping = config.type_mapping.get(btype, TypeMapping())
            for i, cat in enumerate(config.impact_categories.categories):
                query_name = f"{pascal_case(btype)}{pascal_case(cat)}Count"
                desc = (
                    f"Number of {btype.lower()} buildings with {cat.lower()} flooding"
                )
                long_name = f"{btype} with {cat.lower()} flooding (#)"
                filter_str = get_filter(
                    type_mapping=type_mapping,
                    cat_field=config.impact_categories.field,
                    cat_idx=i,
                    bins=config.impact_categories.bins,
                    base_filt=base_filt,
                )
                building_queries.append(
                    MetricModel(
                        name=query_name,
                        select="COUNT(*)",
                        filter=filter_str,
                        description=desc,
                        long_name=long_name,
                    )
                )
        self.infographics_metrics_event.extend(building_queries)
        self.infographics_config["buildings"] = (
            self._make_infographics_config_buildings(config)
        )

    def _setup_svi(self, config: HomesInfographicModel, base_filt) -> None:
        """
        Configure SVI metrics and configuration for infographics.

        Parameters
        ----------
        config : SviInfographicModel
            SVI infographic configuration.
        base_filt : str
            Base SQL filter to apply.
        """
        # Generate queries for all SVI categories and vulnerability levels
        svi_queries = []
        cat_field = config.impact_categories.field
        bins = config.impact_categories.bins
        if config.svi is not None:
            svi_thresholds = config.svi.thresholds
            svi_classes = config.svi.classes
        else:
            svi_thresholds = []
            svi_classes = []
        mapping = config.mapping

        for i, cat in enumerate(config.impact_categories.categories):
            for j, svi_class in enumerate(svi_classes):
                # Build SVI condition based on thresholds
                if j == 0:
                    # First class: SVI < first_threshold
                    svi_cond = f"`SVI` < {svi_thresholds[0]}"
                elif j == len(svi_classes) - 1:
                    # Last class: SVI >= last_threshold
                    svi_cond = f"`SVI` >= {svi_thresholds[-1]}"
                else:
                    # Middle classes: previous_threshold <= SVI < current_threshold
                    svi_cond = f"`SVI` >= {svi_thresholds[j-1]} AND `SVI` < {svi_thresholds[j]}"

                # Build impact category condition based on bins
                if i == 0:
                    # First category: field <= first_bin
                    cat_cond = f"`{cat_field}` <= {bins[0]}"
                elif i == len(config.impact_categories.categories) - 1:
                    # Last category: field > last_bin
                    cat_cond = f"`{cat_field}` > {bins[-1]}"
                else:
                    # Middle categories: previous_bin < field <= current_bin
                    cat_cond = (
                        f"`{cat_field}` > {bins[i-1]} AND `{cat_field}` <= {bins[i]}"
                    )

                # Build type filter using TypeMapping
                type_cond = mapping.to_sql_filter()

                filter_str = combine_filters(type_cond, cat_cond, svi_cond, base_filt)

                name = f"{cat}{svi_class.replace(' ', '')}Vulnerability"
                desc = f"Number of {cat.lower()} homes with {svi_class.lower()} vulnerability"
                long_name = f"{cat} Homes - {svi_class} Vulnerability (#)"

                svi_queries.append(
                    MetricModel(
                        name=name,
                        select="COUNT(*)",
                        filter=filter_str,
                        description=desc,
                        long_name=long_name,
                        show_in_metrics_table=True,
                    )
                )
        self.infographics_metrics_event.extend(svi_queries)
        self.infographics_config["svi"] = self._make_infographics_config_svi(config)

    def _setup_roads(self, config: RoadsInfographicModel) -> None:
        """
        Configure roads metrics and configuration for infographics.

        Parameters
        ----------
        config : RoadsInfographicModel
            Roads infographic configuration.
        """
        # Generate queries for all road categories
        road_queries = []
        cat_field = config.field
        thresholds = config.thresholds
        road_length_field = config.road_length_field

        if config.unit == "meters":
            unit_conversion = 1 / 1000
            unit = "Kilometers"
        elif config.unit == "feet":
            unit_conversion = 1 / 5280
            unit = "Miles"

        for i, cat in enumerate(config.categories):
            name = f"{pascal_case(cat)}FloodedRoadsLength"
            desc = f"{unit} of roads disrupted for {config.users[i].lower()}"
            long_name = f"Length of roads with {cat.lower()} flooding ({unit})"
            select = f"SUM(`{road_length_field}`)*{unit_conversion}"
            filter_str = f"`{cat_field}` >= {thresholds[i]}"
            road_queries.append(
                MetricModel(
                    name=name,
                    description=desc,
                    long_name=long_name,
                    select=select,
                    filter=filter_str,
                    show_in_metrics_table=True,
                )
            )
        self.infographics_metrics_event.extend(road_queries)
        self.infographics_config["roads"] = self._make_infographics_config_roads(config)

    @staticmethod
    def _make_infographics_config_buildings(
        buildings_config: BuildingsInfographicModel,
    ) -> Dict[str, Any]:
        """
        Create infographics configuration dictionary for buildings.

        Parameters
        ----------
        buildings_config : BuildingsInfographicModel
            Building infographic configuration.

        Returns
        -------
        Dict[str, Any]
            Configuration dictionary for building infographics.
        """
        image_path = "{image_path}"
        # Default plot configuration, matching your existing template:
        # Dynamically generate the Info text based on the number of categories and bins
        info_lines = [f"{buildings_config.impact_categories.field}:<br>"]
        for idx, cat in enumerate(buildings_config.impact_categories.categories):
            if idx < len(buildings_config.impact_categories.bins):
                info_lines.append(
                    f"    {cat}: <={buildings_config.impact_categories.bins[idx]} {buildings_config.impact_categories.unit}<br>"
                )
            else:
                # Last category: greater than last bin
                info_lines.append(
                    f"    {cat}: >{buildings_config.impact_categories.bins[-1]} {buildings_config.impact_categories.unit}<br>"
                )

        other_config: Dict[str, Any] = {
            "Plot": {
                "image_scale": 0.15,
                "numbers_font": 20,
                "height": 350,
                "width": 1200,
            },
            "Title": {"text": "Building Impacts", "font": 30},
            "Subtitle": {"font": 25},
            "Legend": {"font": 20},
            "Info": {
                "text": "".join(info_lines),
                "image": "https://openclipart.org/image/800px/302413",
                "scale": 0.1,
            },
        }

        cfg: Dict[str, Any] = {
            "Charts": {},
            "Categories": {},
            "Slices": {},
            "Other": other_config,
        }

        # Categories block - keys have no special characters, Names are original
        if buildings_config.impact_categories.colors is not None:
            for k, cat in enumerate(buildings_config.impact_categories.categories):
                clean_cat_key = pascal_case(cat.replace(" ", "").replace("-", ""))
                cfg["Categories"][clean_cat_key] = {
                    "Name": cat,  # Original name
                    "Color": buildings_config.impact_categories.colors[k],
                }

        # Charts block - keys have no special characters, Names are original
        for i, btype in enumerate(buildings_config.types):
            clean_btype_key = pascal_case(btype.replace(" ", "").replace("-", ""))
            cfg["Charts"][clean_btype_key] = {
                "Name": btype,  # Original name
                "Image": f"{image_path}/{buildings_config.icons[i]}.png",
            }

        # Slices block - reference Chart and Category Names (not keys)
        for i, btype in enumerate(buildings_config.types):
            for cat in buildings_config.impact_categories.categories:
                clean_cat_key = pascal_case(cat.replace(" ", "").replace("-", ""))
                clean_btype_key = pascal_case(btype.replace(" ", "").replace("-", ""))
                slice_key = f"{clean_cat_key}{clean_btype_key}"
                cfg["Slices"][slice_key] = {
                    "Name": f"{cat} {btype}",
                    "Query": f"{pascal_case(btype)}{pascal_case(cat)}Count",
                    "Chart": btype,  # Reference Chart Name, not key
                    "Category": cat,  # Reference Category Name, not key
                }

        return cfg

    @staticmethod
    def _make_infographics_config_svi(
        svi_config: HomesInfographicModel,
    ) -> Dict[str, Any]:
        """
        Create infographics configuration dictionary for SVI.

        Parameters
        ----------
        svi_config : SviInfographicModel
            SVI infographic configuration.

        Returns
        -------
        Dict[str, Any]
            Configuration dictionary for SVI infographics.
        """
        image_path = "{image_path}"
        svi_classes = svi_config.svi.classes if svi_config.svi is not None else []
        charts = {}
        slices = {}
        categories_cfg = {}

        # Charts block - keys have no special characters, Names are original
        for cat in svi_config.impact_categories.categories:
            clean_cat_key = pascal_case(cat.replace(" ", "").replace("-", ""))
            charts[clean_cat_key] = {
                "Name": cat,  # Original name
                "Image": f"{image_path}/house.png",
            }

        # Categories block - keys have no special characters, Names are original
        for idx, svi_class in enumerate(svi_classes):
            color = (
                svi_config.svi.colors[idx] if svi_config.svi is not None else "#cccccc"
            )
            clean_svi_key = (
                pascal_case(svi_class.replace(" ", "").replace("-", ""))
                + "Vulnerability"
            )
            categories_cfg[clean_svi_key] = {
                "Name": f"{svi_class} Vulnerability",  # Original name with "Vulnerability"
                "Color": color,
            }

        # Slices block - reference Chart and Category Names (not keys)
        for cat in svi_config.impact_categories.categories:
            for svi_class in svi_classes:
                clean_cat_key = pascal_case(cat.replace(" ", "").replace("-", ""))
                clean_svi_key = (
                    pascal_case(svi_class.replace(" ", "").replace("-", ""))
                    + "Vulnerability"
                )
                slice_key = f"{clean_cat_key}{clean_svi_key}"
                name = f"{cat} {svi_class.lower()} vulnerability homes"
                query = (
                    f"{cat}{svi_class.replace(' ', '').replace('-', '')}Vulnerability"
                )
                slices[slice_key] = {
                    "Name": name,
                    "Query": query,
                    "Chart": cat,  # Reference Chart Name, not key
                    "Category": f"{svi_class} Vulnerability",  # Reference Category Name, not key
                }

        # Info text - dynamically build threshold information
        bins = svi_config.impact_categories.bins
        unit = svi_config.impact_categories.unit
        thresholds = svi_config.svi.thresholds if svi_config.svi is not None else []

        info_lines = [
            "Thresholds:<br>",
        ]

        # Add impact category threshold information
        for i, cat in enumerate(svi_config.impact_categories.categories):
            if i == 0:
                info_lines.append(f"    {cat}: <= {bins[0]} {unit}<br>")
            elif i == len(svi_config.impact_categories.categories) - 1:
                info_lines.append(f"    {cat}: > {bins[-1]} {unit}<br>")
            else:
                info_lines.append(f"    {cat}: {bins[i-1]} - {bins[i]} {unit}<br>")

        info_lines.append("<br>SVI Classes:<br>")

        # Add SVI threshold information
        for i, svi_class in enumerate(svi_classes):
            if i == 0:
                info_lines.append(f"    {svi_class}: < {thresholds[0]}<br>")
            elif i == len(svi_classes) - 1:
                info_lines.append(f"    {svi_class}: >= {thresholds[-1]}<br>")
            else:
                info_lines.append(
                    f"    {svi_class}: {thresholds[i-1]} - {thresholds[i]}<br>"
                )

        info_lines.extend(
            [
                "'Since some homes do not have an SVI,<br>",
                "total number of homes might be different <br>",
                "between this and the above graph.'",
            ]
        )

        other_config = {
            "Plot": {
                "image_scale": 0.15,
                "numbers_font": 20,
                "height": 350,
                "width": 600,
            },
            "Title": {"text": "Impacted Homes", "font": 30},
            "Subtitle": {"font": 25},
            "Legend": {"font": 20},
            "Info": {
                "text": "".join(info_lines),
                "image": "https://openclipart.org/image/800px/302413",
                "scale": 0.1,
            },
        }
        cfg = {
            "Charts": charts,
            "Categories": categories_cfg,
            "Slices": slices,
            "Other": other_config,
        }
        return cfg

    @staticmethod
    def _make_infographics_config_roads(
        roads_config: RoadsInfographicModel,
    ) -> Dict[str, Any]:
        """
        Create infographics configuration dictionary for roads.

        Parameters
        ----------
        roads_config : RoadsInfographicModel
            Roads infographic configuration.

        Returns
        -------
        Dict[str, Any]
            Configuration dictionary for roads infographics.
        """
        image_path = "{image_path}"

        # Charts block - key has no special characters, Name is original
        chart_key = "FloodedRoads"
        chart_name = "Flooded Roads"
        charts = {chart_key: {"Name": chart_name}}

        # Categories block - keys have no special characters, Names are original
        categories_cfg = {}
        for idx, cat in enumerate(roads_config.categories):
            clean_cat_key = (
                pascal_case(cat.replace(" ", "").replace("-", "")) + "Flooding"
            )
            cat_name = f"{cat} Flooding"
            categories_cfg[clean_cat_key] = {
                "Name": cat_name,  # Original name with "Flooding"
                "Color": roads_config.colors[idx],
                "Image": f"{image_path}/{roads_config.icons[idx]}.png",
            }

        # Slices block - reference Chart and Category Names (not keys)
        slices = {}
        for idx, cat in enumerate(roads_config.categories):
            clean_cat_key = (
                pascal_case(cat.replace(" ", "").replace("-", "")) + "Flooding"
            )
            cat_name = f"{cat} Flooding"
            query_name = f"{pascal_case(cat)}FloodedRoadsLength"
            slices[clean_cat_key] = {
                "Name": cat_name,
                "Query": query_name,
                "Chart": chart_name,  # Reference Chart Name, not key
                "Category": cat_name,  # Reference Category Name, not key
            }

        # Info text
        thresholds = roads_config.thresholds
        # Consistent unit naming
        if roads_config.unit == "feet":
            unit = "Miles"
        else:
            unit = "Kilometers"
        info_lines = ["Thresholds:<br>"]
        for user, threshold in zip(roads_config.users, thresholds):
            info_lines.append(f"    {user}: {threshold} {roads_config.unit}<br>")
        other_config = {
            "Plot": {
                "image_scale": 0.1,
                "numbers_font": 20,
                "height": 350,
                "width": 600,
            },
            "Title": {"text": "Interrupted roads", "font": 30},
            "Subtitle": {"font": 25},
            "Y_axis_title": {"text": unit},
            "Info": {
                "text": "".join(info_lines),
                "image": f"{image_path}/info.png",
                "scale": 0.1,
            },
        }
        cfg = {
            "Charts": charts,
            "Categories": categories_cfg,
            "Slices": slices,
            "Other": other_config,
        }
        return cfg

    def _make_infographics_config_risk(self, risk_config: RiskInfographicModel) -> dict:
        """
        Create infographics configuration dictionary for risk infographics.

        Parameters
        ----------
        risk_config : RiskInfographicModel
            Risk infographic configuration.

        Returns
        -------
        dict
            Configuration dictionary for risk infographics.
        """
        image_path = "{image_path}"
        rps = self.return_periods
        homes = risk_config.homes
        svi = homes.svi if homes.svi is not None else None

        # Charts block - keys have no special characters, Names are original
        charts = {}
        for rp in rps:
            chart_key = f"RP{int(rp)}Y"
            chart_name = f"{int(rp)}Y"
            charts[chart_key] = {
                "Name": chart_name,  # Original name
                "Image": f"{image_path}/house.png",
            }

        # Categories block - keys have no special characters, Names are original
        categories = {}
        if svi:
            for idx, svi_class in enumerate(svi.classes):
                cat_key = (
                    pascal_case(svi_class.replace(" ", "").replace("-", ""))
                    + "Vulnerability"
                )
                cat_name = f"{svi_class} Vulnerability"
                categories[cat_key] = {
                    "Name": cat_name,  # Original name with "Vulnerability"
                    "Color": svi.colors[idx] if svi.colors else "#cccccc",
                }
        else:
            # Create a single "All Homes" category when no SVI data
            categories["AllHomes"] = {
                "Name": "All Homes",
                "Color": "#88A2AA",  # Default color
            }

        # Slices block - reference Chart and Category Names (not keys)
        slices = {}
        if svi:
            for rp in rps:
                chart_name = f"{int(rp)}Y"
                for idx, svi_class in enumerate(svi.classes):
                    clean_svi = svi_class.replace(" ", "").replace("-", "")
                    cat_name = f"{svi_class} Vulnerability"
                    slice_key = f"{pascal_case(svi_class.replace(' ', '').replace('-', ''))}VulnerabilityRP{int(rp)}Y"
                    name = f"{int(rp)}Y {svi_class} Vulnerability"
                    query = f"ImpactedHomes{int(rp)}Y{clean_svi}SVI"
                    slices[slice_key] = {
                        "Name": name,
                        "Query": query,
                        "Chart": chart_name,  # Reference Chart Name, not key
                        "Category": cat_name,  # Reference Category Name, not key
                    }
        else:
            # Create slices for each return period without SVI breakdown
            for rp in rps:
                chart_name = f"{int(rp)}Y"
                slice_key = f"AllHomesRP{int(rp)}Y"
                name = f"{int(rp)}Y All Homes"
                query = f"ImpactedHomes{int(rp)}Y"
                slices[slice_key] = {
                    "Name": name,
                    "Query": query,
                    "Chart": chart_name,  # Reference Chart Name, not key
                    "Category": "All Homes",  # Reference Category Name, not key
                }

        # Other block: static info, but use config where possible
        other = {
            "Expected_Damages": {
                "title": "Expected annual damages",
                "query": "ExpectedAnnualDamages",
                "image": f"{image_path}/money.png",
                "image_scale": 1.3,
                "title_font_size": 25,
                "numbers_font_size": 20,
                "height": 300,
            },
            "Flooded": {
                "title": f"Number of homes with a high chance of being flooded in a {risk_config.flood_exceedance.period}-year period",
                "query": "LikelyFloodedHomes",
                "image": f"{image_path}/house.png",
                "image_scale": 0.7,
                "title_font_size": 25,
                "numbers_font_size": 20,
                "height": 300,
            },
            "Return_Periods": {
                "title": "Buildings impacted",
                "font_size": 25,
                "image_scale": 0.2,
                "numbers_font": 15,
                "subtitle_font": 22,
                "legend_font": 20,
                "plot_height": 300,
            },
            "Info": {
                "text": (
                    "Thresholds:<br>"
                    f"    Impacted: >= {risk_config.flood_exceedance.threshold} {risk_config.flood_exceedance.unit}<br>"
                ),
                "image": "https://openclipart.org/image/800px/302413",
                "scale": 0.1,
            },
        }

        cfg = {
            "Charts": charts,
            "Categories": categories,
            "Slices": slices,
            "Other": other,
        }
        self.infographics_config["risk"] = cfg
        return cfg
